#!/usr/bin/env python3
'''
osci deploy helpers
'''

import logging
import os
import random
import subprocess
import string
import yaml

import keystoneauth1.identity.v3 as keystoneauth1
import keystoneauth1.session as keystoneauth1_session
from neutronclient.v2_0 import client as ne_client
from jinja2 import Environment, FileSystemLoader


JM_NAME_TEMPL = 'osci-master/{}'

SLAVE_NAME_TEMPLATES = {
    'slave': 'osci-slave/{}',  # jsv
    'lab': 'osci-lab/{}',      # jlb
    'task': 'osci-task/{}'     # jtk
}


def rnd_str(length=32):
    '''Return lower-case alpha-numeric random string, a la pwgen.

    :param length: number of characters to generate
    :returns: lower-case alpha-numeric random string
    '''
    _str = ''
    for _ in range(length):
        _str += random.choice(string.lowercase + string.digits)
    return _str


def read_yaml(the_file):
    '''Returns yaml data from provided file name

    :param the_file: yaml file name to read
    :returns: dictionary of yaml data from file
    '''
    if not os.path.exists(the_file):
        raise ValueError('File not found: {}'.format(the_file))
    with open(the_file) as yaml_file:
        data = yaml.safe_load(yaml_file.read())
    return data


def write_yaml(data, the_file):
    '''Save yaml data dictionary to a yaml file

    :param the_file: yaml file name to write
    :returns: dictionary of yaml data to write to file
    '''
    with open(the_file, 'w') as yaml_file:
        yaml_file.write(yaml.dump(data, default_flow_style=False))


def safe_mkdir(directory):
    '''Make directory if it doesn't exist.  a la mkdir -p
    '''
    if not os.path.exists(directory):
        os.makedirs(directory)


def render(source, target, context, templates_dir=None):
    '''Render a template.

       The `source` path, if not absolute, is relative to the `templates_dir`.

       The `target` path should be absolute.

       The context should be a dict containing the values to be replaced in
       the template.

       If omitted, `templates_dir` defaults to the current working dir.
    '''

    if templates_dir is None:
        templates_dir = os.getcwd()
    loader = Environment(loader=FileSystemLoader(templates_dir))
    try:
        source = source
        template = loader.get_template(source)
    except Exception:
        logging.error('Could not load template {} from {}.'.format(
            source, templates_dir))

    safe_mkdir(templates_dir)
    safe_mkdir(os.path.dirname(target))
    with open(target, 'w') as _file:
        _file.write(template.render(context))


def get_juju_application_status(application):
    '''Return juju application yaml status as dict.'''
    return yaml.load(subprocess.check_output(
                     ['juju', 'status', application,
                      '--format', 'yaml']))


def copy_file_to_juju_unit(local_file, remote_file, juju_unit, owner="jenkins",
                           group="jenkins", perms="640"):
    '''Copy a local file to a juju unit and set its ownership and permissions.
    '''

    logging.info('Copying {} to {}:{}'.format(
        local_file, juju_unit, remote_file))

    subprocess.check_output(
        ['juju', 'scp', local_file, '{}:{}'.format(juju_unit, remote_file)])

    subprocess.check_output(
        ['juju', 'ssh', juju_unit,
         'chown', '{}:{}'.format(owner, group), remote_file])

    subprocess.check_output(
        ['juju', 'ssh', juju_unit,
         'chmod', perms, remote_file])


def copy_file_on_juju_unit(src_file, dest_file, juju_unit, owner="jenkins",
                           group="jenkins", perms="640"):
    '''Copy a remote file from one location to another location on a juju unit
       and set its ownership and permissions.
    '''

    logging.info('Copying {} to {} on {}'.format(
        src_file, dest_file, juju_unit))

    subprocess.check_output(
        ['juju', 'ssh', juju_unit, 'sudo',
         'cp', '-fvp', src_file, dest_file])

    subprocess.check_output(
        ['juju', 'ssh', juju_unit, 'sudo',
         'chown', '{}:{}'.format(owner, group), dest_file])

    subprocess.check_output(
        ['juju', 'ssh', juju_unit, 'sudo',
         'chmod', perms, dest_file])


# Old dup, deprecate?
def XXX_get_num_units(yaml_data, outer, service):
    '''Return num_units from service in yaml data from juju bundle file

    :param yaml_data: yaml data from juju bundle file, as dict
    :param outer: first, outer key in juju bundle yaml dict data
    :param service: juju service to query for num units
    :returns: value of num_units from juju bundle yaml dict data
    '''
    logging.debug('yaml data: {}'.format(yaml_data))
    logging.debug('outer: {}'.format(outer))
    logging.debug('service: {}'.format(service))
    return yaml_data[outer]['services'][service]['num_units']


def recursive_dict_key_search(data, key):
    '''Recursively look for key match in dictionaries, return value.

    :param data: dictionary data
    :param key: key to find
    :returns: value of found key
    '''
    if key in data:
        return data[key]
    for _, val in data.items():
        if isinstance(val, dict):
            ret = recursive_dict_key_search(val, key)
            if ret is not None:
                return ret


def validate_config(data, expected):
    '''Very light check for existence of non-empty data

    :param data: dictionary of k:v pairs
    :param expected: list of items to be validated as key in data
    :returns: None if found as expected, key if not found or key
        value is None
    '''
    for k in expected:
        val = recursive_dict_key_search(data, k)
        if val != '' and val != 'null' and val is not None:
            logging.debug('found {}: {}'.format(k, val))
        else:
            return k


def generate_list_of_cidrs(network, octet, quantity, bits):
    '''Create a list of IPV4 networks in CIDR notation, generated by
    incrementing a specified octet, the specified number of times.

    :param network: network string (eg, 192.168.7.0 or 10.200.0.0, non-CIDR)
    :param octet: octet to increment (eg, 1.2.3.4)
    :param quantity: qty of networks to generate
    :param bits: bits to fill, ie. the slash from CIDR notation
    :returns: list of network string items in CIDR format
    '''
    nets = []
    logging.debug('operating on: {}'.format(network.split('.')))
    this_addr = ''
    for i in range(0, quantity):
        octs = network.split('.')
        oct_new = int(octs[octet - 1]) + i
        if oct_new > 254:
            raise ValueError('IPv4 octec decimal cannot exceed 254! {} was '
                             'attempted.'.format(oct_new))
        octs[octet - 1] = str(oct_new)
        this_addr = '.'.join(octs)
        nets.append('{}/{}'.format(this_addr, bits))
    logging.debug('networks to build ({}): {}'.format(len(nets), nets))
    return nets


def predict_juju_unit_names(conf_mod, jsv_qty, jtk_qty, jlb_qty):
    '''Create a list of juju jenkins master and slave service
       units, predictively.
    '''
    units = []
    units.append(JM_NAME_TEMPL.format('0'))

    for _templ_qty_map in [(SLAVE_NAME_TEMPLATES['slave'], jsv_qty),
                           (SLAVE_NAME_TEMPLATES['task'], jtk_qty),
                           (SLAVE_NAME_TEMPLATES['lab'], jlb_qty)]:
        _templ = _templ_qty_map[0]
        _qty = _templ_qty_map[1]
        for unit in range(0, _qty):
            units.append(_templ.format(unit))

    return units


def predict_model_names(conf_mod, jsv_qty, jtk_qty, jlb_qty):
    '''Return a list of juju model names to create'''
    models = []
    models.append("{}-{}{}".format(conf_mod['model-name-base'],
                                   conf_mod['master-abbrv'],
                                   0))

    for _abbrv_qty_map in [(conf_mod['slave-abbrv'], jsv_qty),
                           (conf_mod['task-abbrv'], jtk_qty),
                           (conf_mod['lab-abbrv'], jlb_qty)]:
        _abbrv = _abbrv_qty_map[0]
        _qty = _abbrv_qty_map[1]
        for unit in range(0, _qty):
            models.append("{}-{}{}".format(conf_mod['model-name-base'],
                                           _abbrv,
                                           str(unit).zfill(2)))
    return models


# Unused, deprecate?:
def XXX_floating_ip_is_available(neutron):
    '''Check if a floating ip is available to the tenant, using the
    provided neutron client object.

    :param neutron: neutron client object
    :returns: True if a floating IP is available to the tenant
    '''
    ips = neutron.list_floatingips(status='DOWN')['floatingips']

    if len(ips) > 0:
        logging.debug('floating ips available to tenant: {}'.format(ips))
        return True
    else:
        return False


# Unused, deprecate?:
def XXX_has_floating_ip(addresses, fip_str):
    '''Inspect list of addresses for an item which starts with the specified
    string.  Returns the (first) matching item found, returns None if not.

    :param addresses: list of dicts containing nova addresses data
    :param fip_str: floating IP string to match, generally first 2 or 3 octets
    '''
    for addr in addresses:
        if addr['addr'].startswith(fip_str):
            return addr['addr']
    return None


# Unused, deprecate?:
def XXX_f_contains_str(the_file, str_find):
    '''Search a file for a string

    :param the_file: file name to read
    :param str_find: string to find in the file
    :returns: True if string is found in file
    '''
    _pos = open(the_file, 'r').read().find(str_find)
    if _pos >= 0:
        return True
    else:
        return False


# Unused, deprecate?:
def XXX_replace_line_with_lines(file_name, str_find, with_lines):
    '''Inject lines into file, where string is found in exactly and
       only one line.  Return None if successful, return a message otherwise.

    :param file_name: file name as string, to read and write
    :param str_find: string to find in a line
    :param str_lines: list of lines to inject
    '''
    with open(file_name, 'r') as r_file:
        lines = r_file.readlines()

    indices = []
    for idx, line in enumerate(lines):
        if str_find in line:
            indices.append(idx)
    if len(indices) > 1:
        return 'Could not replace line, found more than one match.'
    elif len(indices) == 0:
        return 'Could not replace line, no match in {}'.format(file_name)
    elif len(indices) == 1:
        logging.info('string find for {}: {}'.format(str_find, indices))
        with open(file_name, 'w') as w_file:
            out_lines = (lines[0:indices[0]] +
                         with_lines +
                         lines[indices[0] + 1:])
            w_file.writelines(out_lines)


def dict_to_yaml(dict_data):
    return yaml.dump(dict_data, default_flow_style=False)


def get_keystone_v3_session():
    '''Get a keystone v3 session based on env vars.'''

    auth = keystoneauth1.Password(
        username=os.environ['OS_USERNAME'],
        user_domain_name=os.environ['OS_USER_DOMAIN_NAME'],
        password=os.environ['OS_PASSWORD'],
        project_name=os.environ['OS_PROJECT_NAME'],
        project_domain_name=os.environ['OS_PROJECT_DOMAIN_NAME'],
        auth_url=os.environ['OS_AUTH_URL']
    )

    return keystoneauth1_session.Session(auth=auth)


def get_neutron_client(session):
    '''Get a neutron client connection based on ksv3 session and env vars.'''
    return ne_client.Client(session=session,
                            region_name=os.environ['OS_REGION_NAME'])
