#!/usr/bin/env python
'''
Creates networks, subnets and prepares info for later consumption by
osci jenkins slaves regarding juju models.
'''

import logging
import yaml
from netaddr import IPNetwork
from lib.osci_deploy_helpers import (
    copy_file_to_juju_unit,
    copy_file_on_juju_unit,
    generate_list_of_cidrs,
    get_juju_application_status,
    get_neutron_client,
    get_keystone_v3_session,
    predict_juju_unit_names,
    predict_model_names,
    read_yaml,
    render,
    safe_mkdir,
    validate_config,
    write_yaml
)

# Set default logging level
logging.basicConfig(level=logging.DEBUG)

EV = 'environments'
DEPLOY_CONFIG_FILE = 'config/nets-and-models.yaml'


def validate_network_config(conf):
    '''Confirm network config, verify cidr format using netaddr,
       check for sizing logic with regards to incrementation.
    '''
    ip_cidr = IPNetwork(conf['cidr'])
    ip_bits = int(ip_cidr.prefixlen)
    ret = ''

    if conf['cidr'] != str(ip_cidr.cidr):
        ret += ('Expected CIDR notation, ie. 10.100.0.0/16 but '
                'found: {}'.format(conf['cidr']))
    if ip_bits > 24:
        logging.warn('That may not be enough IP space ({}). '
                     'Proceeding anyway.'.format(ip_cidr))
    if int(conf['var-octet']) < 1 or int(conf['var-octet']) > 4:
        ret += ('Octet out of range: {}. '
                '1-4 are valid. '.format(conf['var-octet']))
    if conf['var-octet'] == '1':
        logging.warn('The first octet should not be targeted, unless '
                     'you have that public ip space.  Proceeding anyway.')
    if conf['var-octet'] == '2' and ip_bits < 16:
        ret += ('Can\'t split lower than /16 '
                'while incrementing 2nd octet!  ')
    if conf['var-octet'] == '3' and ip_bits < 24:
        ret += ('Can\'t split lower than /24 '
                'while incrementing 2nd octet!  ')
    if conf['var-octet'] == '4':
        ret += 'Can\'t use 4th octet.  Recommend 2nd.'

    logging.debug('ip network: {}'.format(ip_cidr.network))
    logging.debug('ip prefixlen: {}'.format(ip_cidr.prefixlen))
    return ret


def build_nets_and_subs(neutron_c, d_models_data, conf):
    '''Build new neutron subnets and networks from dict data, and connect
       to neutron router.'''
    conf_net = conf['osci']['network']
    conf_mod = conf['osci']['model']

    base_name = conf_mod['model-name-base']
    var_oct = conf_net['var-octet']
    router_name = conf_net['neutron-router']
    dns = conf_net['neutron-dns']

    for k, v in d_models_data['models'].items():
        net_str = v['network']
        sub_name = "{}_{}_{}".format(base_name,
                                     net_str.split('.')[int(var_oct) - 1],
                                     'sub')
        net_name = "{}_{}_{}".format(base_name,
                                     net_str.split('.')[int(var_oct) - 1],
                                     'net')

        # Create neutron network
        networks = neutron_c.list_networks(name=net_name)
        if len(networks['networks']) == 0:
            logging.info('Creating network: %s',
                         net_name)
            network_msg = {
                'network': {
                    'name': net_name,
                    'shared': False
                }
            }
            neutron_net = neutron_c.create_network(network_msg)['network']
        else:
            logging.info('Network %s already exists.', net_name)
            neutron_net = networks['networks'][0]

        # Check if subnet name and IP space is free
        subnets = neutron_c.list_subnets(name=sub_name)
        if len(subnets['subnets']) == 0:
            logging.info('Creating subnet: %s', sub_name)
            this_cidr = IPNetwork(net_str)
            alloc_start = this_cidr[2]
            alloc_end = this_cidr[-57]  # leave some static space available

            subnet_msg = {
                'subnet': {
                    'name': sub_name,
                    'network_id': neutron_net['id'],
                    'enable_dhcp': True,
                    'cidr': str(this_cidr),
                    'ip_version': 4,
                    'dns_nameservers': dns.split(',')
                }
            }
            subnet_msg['subnet']['allocation_pools'] = [{
                'start': str(alloc_start),
                'end': str(alloc_end)
            }]
            logging.info(subnet_msg)
            neutron_sub = neutron_c.create_subnet(subnet_msg)['subnet']
        else:
            logging.info('Subnet %s already exists.', net_name)
            neutron_sub = subnets['subnets'][0]

        # Wire the network to the router
        routers = neutron_c.list_routers(name=router_name)
        if len(routers['routers']) == 0:
            raise ValueError('Unable to locate router')
        else:
            ports = neutron_c.list_ports(
                device_owner='network:router_interface',
                network_id=neutron_net['id']
            )
            if len(ports['ports']) == 0:
                logging.info('Adding interface from %s to %s',
                             router_name, sub_name)
                neutron_rtr = routers['routers'][0]
                neutron_c.add_interface_router(
                    neutron_rtr['id'],
                    {'subnet_id': neutron_sub['id']}
                )
            else:
                logging.info('Router already connected to subnet %s', sub_name)

        d_models_data['models'][k]['network-id'] = str(neutron_net['id'])


def get_num_units():
    '''Return specific juju application unit quantities from juju model status.
    '''
    jm_stat = get_juju_application_status('osci-master')
    jsv_stat = get_juju_application_status('osci-slave')
    jtk_stat = get_juju_application_status('osci-task')
    jlb_stat = get_juju_application_status('osci-lab')

    return (len(jm_stat['applications']['osci-master']['units']),
            len(jsv_stat['applications']['osci-slave']['units']),
            len(jtk_stat['applications']['osci-task']['units']),
            len(jlb_stat['applications']['osci-lab']['units']))


def generate_models_data(conf):
    '''Read and validate configs, build networks, dump models:networks yaml.
    '''
    conf = read_yaml(DEPLOY_CONFIG_FILE)

    # Basic non-null general parameter validation
    expected = ['model-name-base', 'master-abbrv', 'slave-abbrv', 'split',
                'out-dir']
    ret = validate_config(conf['osci'], expected)
    if ret:
        logging.debug('conf: {}'.format(conf['osci']))
        raise ValueError('Config data missing. {}'.format(ret))

    # Break out config parameters
    conf_net = conf['osci']['network']
    conf_mod = conf['osci']['model']
    net_split = conf['osci']['network']['split']

    # Predictively construct new juju model names from unit names
    jma_num_units, jsv_num_units, jtk_num_units, jlb_num_units = \
        get_num_units()

    junits = predict_juju_unit_names(conf_mod=conf_mod,
                                     jsv_qty=jsv_num_units,
                                     jtk_qty=jtk_num_units,
                                     jlb_qty=jlb_num_units)

    models = predict_model_names(conf_mod=conf_mod,
                                 jsv_qty=jsv_num_units,
                                 jtk_qty=jtk_num_units,
                                 jlb_qty=jlb_num_units)

    logging.info('Models: {}'.format(models))

    # Network separation
    if not net_split:
        logging.info('No network separation.')
        nets = [None] * len(junits)
    else:
        logging.info('Using layer 2 network separation.')

        # Basic non-null network parameter validation
        expected = ['cidr', 'var-octet', 'neutron-router',
                    'neutron-net', 'neutron-sub', 'neutron-dns']
        ret = validate_config(conf_net, expected)
        if ret:
            raise ValueError('Network config data missing. {}'.format(ret))

        # Validate IP as CIDR;  Validate network size and incrementating logic
        ret = validate_network_config(conf=conf_net)
        if ret:
            raise ValueError('IP, network or logic issue(s): {}'.format(ret))

        # Construct new network config data list
        ip_cidr = IPNetwork(conf_net['cidr'])
        total_units = (jma_num_units + jsv_num_units +
                       jtk_num_units + jlb_num_units)
        nets = generate_list_of_cidrs(network=str(ip_cidr.network),
                                      bits=int(ip_cidr.prefixlen),
                                      octet=int(conf_net['var-octet']),
                                      quantity=total_units)

    # Ensure lists are the same length & smash into a dictionary
    check_len = [len(models), len(nets), len(junits)]
    if check_len != [check_len[0]] * 3:
        logging.debug('models ({}): {}'.format(len(models), models))
        logging.debug('nets ({}): {}'.format(len(nets), nets))
        logging.debug('junits ({}): {}'.format(len(junits), junits))
        raise ValueError('Model, network and juju unit counts should be equal')
    d_models_data = {
        'models': {},
        'units': {},
    }
    for idx, model in enumerate(models):
        d_models_data['models'][model] = {
            'network': nets[idx],
            'unit': junits[idx]
        }

        d_models_data['units'][junits[idx]] = {
            'network': nets[idx],
            'model': model
        }

    return d_models_data


def write_models_data_files(conf, d_models_data):
    '''Write out yaml files containing models, units, networks data
       for consumption by osci jenkins slaves.
    '''
    conf_other = conf['osci']['other']

    logging.debug('New model and network predictions: {}'.format(
                  yaml.dump(d_models_data, default_flow_style=False)))
    safe_mkdir(conf_other['out-dir'])
    write_yaml(d_models_data,
               '{}/models_data.yaml'.format(conf_other['out-dir']))

    # Generate oscirc files and copy into units
    for model, v in d_models_data['models'].items():
        unit_str = v['unit']
        unit_str_friendly = unit_str.replace('/', '-')

        oscirc_template_data = {
            'osci_controller_name': model,  # Just use model name
            'osci_friendly_name': unit_str_friendly,
            'osci_juju_name': unit_str,
            'osci_model_name': model,
            'osci_net_cidr': v['network'],
            'osci_net_id': v['network-id'],
        }

        logging.info('{}'.format(oscirc_template_data))

        target = '{}/{}/oscirc'.format(conf_other['out-dir'],
                                       unit_str_friendly)

        render('oscirc', target, oscirc_template_data, 'templates')

        copy_file_to_juju_unit(target,
                               '/home/ubuntu/oscirc',
                               unit_str, owner="ubuntu", group="ubuntu",
                               perms="640")

        copy_file_to_juju_unit('/home/ubuntu/novarc',
                               '/home/ubuntu/novarc',
                               unit_str, owner="ubuntu", group="ubuntu",
                               perms="640")

        copy_file_on_juju_unit('/home/ubuntu/novarc',
                               '/var/lib/jenkins/novarc',
                               unit_str, owner="root", group="root",
                               perms="644")

        copy_file_on_juju_unit('/home/ubuntu/oscirc',
                               '/var/lib/jenkins/oscirc',
                               unit_str, owner="root", group="root",
                               perms="644")


def misc_file_ops(conf, d_models_data):
    '''Perform misc file operations on units.'''

    for model, v in d_models_data['models'].items():
        unit_str = v['unit']

        copy_file_on_juju_unit('/home/ubuntu/.bashrc',
                               '/var/lib/jenkins/.bashrc',
                               unit_str, owner="jenkins", group="jenkins",
                               perms="640")

        copy_file_on_juju_unit('/home/ubuntu/.profile',
                               '/var/lib/jenkins/.profile',
                               unit_str, owner="jenkins", group="jenkins",
                               perms="640")


def main():
    '''Main procedure. And, Go!'''

    conf = read_yaml(DEPLOY_CONFIG_FILE)
    d_models_data = generate_models_data(conf)

    # Build networks and subnets if called for by config
    if conf['osci']['network']['split']:
        ksv3_session = get_keystone_v3_session()
        neutron_c = get_neutron_client(ksv3_session)

        build_nets_and_subs(neutron_c,
                            d_models_data,
                            conf)

    write_models_data_files(conf, d_models_data)

    misc_file_ops(conf, d_models_data)


if __name__ == '__main__':
    main()
