#!/usr/bin/env python
'''
OSCI utility functions common to deployment, admin and ongoing use.
'''

import distro_info
import logging
import multiprocessing
import optparse
import os
import random
import paramiko
import socket
import string
import subprocess
import sys
import tempfile
import time
import uuid
import yaml
import bzrlib.errors
from bzrlib import branch
from bzrlib.plugin import load_plugins
from jinja2 import Environment, FileSystemLoader
from launchpadlib.launchpad import Launchpad
from lazr.restfulclient.errors import Unauthorized

# Common constants
MAP_FILE = 'populate/mappings.yaml'
LPL_CRED_FN = os.path.expanduser("~") + "/.config/tarmac/credentials"

try:
    WORKSPACE = os.environ['WORKSPACE']
except KeyError:
    WORKSPACE = tempfile.mkdtemp()

try:
    JENV_NAME = os.environ['JENV_NAME']
except KeyError:
    JENV_NAME = None

try:
    BUILD_NUMBER = os.environ['BUILD_NUMBER']
except KeyError:
    BUILD_NUMBER = '000'


# Helpers
def XXX_rnd_str(length=32):
    '''Return lower-case alpha-numeric random string, a la pwgen.

    :param length: number of characters to generate
    :returns: lower-case alpha-numeric random string
    '''
    _str = ''
    for _ in range(length):
        _str += random.choice(string.lowercase + string.digits)
    return _str


def fill_dict_with_epoch_time(data):
    _d = {}
    for k in data.keys():
        _d[k] = time.time()
    return _d


def invert_dict_of_lists(d):
    """Inverts a dictionary of lists."""
    _inverted = {}
    for k, v in d.iteritems():
        for item in v:
            if item in _inverted.keys():
                _inverted[item].append(k)
            else:
                _inverted[item] = [k]
    return _inverted


def strip_chars(string, chars='+~/\\'):
    '''Strip characters from string'''
    return string.translate(None, chars)


def touch(file_name):
    '''create a file or update its mtime, a la touch
    '''
    with open(file_name, 'a'):
        os.utime(file_name, None)
    if not os.path.exists(file_name):
        raise ValueError('Touched file not found: {}'.format(file_name))


def read_yaml(the_file):
    '''Returns yaml data from provided file name

    :param the_file: yaml file name to read
    :returns: dictionary of yaml data from file
    '''
    if not os.path.exists(the_file):
        raise ValueError('File not found: {}'.format(the_file))
    with open(the_file) as yaml_file:
        data = yaml.safe_load(yaml_file.read())
    return data


def write_yaml(data, the_file):
    '''Save yaml data dictionary to a yaml file

    :param the_file: yaml file name to write
    :returns: dictionary of yaml data to write to file
    '''
    with open(the_file, 'w') as yaml_file:
        yaml_file.write(yaml.dump(data, default_flow_style=False))


def safe_mkdir(directory):
    '''Make directory if it doesn't exist.  a la mkdir -p
    '''
    if not os.path.exists(directory):
        os.makedirs(directory)


def read_file(file_name):
    '''Read file into array of lines with some validation
    '''
    if os.path.isfile(file_name):
        with open(file_name) as _file:
            _lines = _file.readlines()
            logging.debug('Read {} lines from {}'.format(len(_lines),
                                                         file_name))
            return _lines
    else:
        logging.warn('File not found: {}'.format(file_name))
        return False


def juju_stat(jenv_name=None, timeout=10):
    '''Get juju status, return dict data
    '''
    try:
        if jenv_name:
            cmd = ('timeout {} juju status --format=yaml -e '
                   '{}'.format(timeout, jenv_name))
        else:
            cmd = 'timeout {} juju status --format=yaml'.format(timeout)
        j_stat = yaml.load(subprocess.check_output(cmd.split()))
    except subprocess.CalledProcessError:
        logging.info('Juju status attempt returned non-zero, '
                     'most likely no juju enviro bootstrapped.')
        j_stat = {}
    return j_stat


def XXX_get_num_units(yaml_data, outer, service):
    '''Return num_units from service in yaml data from juju bundle file

    :param yaml_data: yaml data from juju bundle file, as dict
    :param outer: first, outer key in juju bundle yaml dict data
    :param service: juju service to query for num units
    :returns: value of num_units from juju bundle yaml dict data
    '''
    logging.debug('yaml data: {}'.format(yaml_data))
    logging.debug('outer: {}'.format(outer))
    logging.debug('service: {}'.format(service))
    return yaml_data[outer]['services'][service]['num_units']


def dict_unicode_to_str(ddata):
    '''Convert elements in a dictionary from unicode to string utf8

    :ddata: dictionary data
    '''
    if isinstance(ddata, dict):
        return {dict_unicode_to_str(key): dict_unicode_to_str(value)
                for key, value in ddata.iteritems()}
    elif isinstance(ddata, unicode):
        return ddata.encode('utf-8')
    elif isinstance(ddata, list):
        return [dict_unicode_to_str(element) for element in ddata]
    else:
        return ddata


def openstack_release(ubuntu_release):
    '''Return the OpenStack release version shipped with a particular
    release of Ubuntu.

    :param ubuntu_release: Ubuntu release codename (eg, utopic)
    :returns: OpenStack release name as string
    '''

    _map = read_yaml(MAP_FILE)
    try:
        return _map['release-map'][ubuntu_release]
    except KeyError:
        logging.error('Could not determine Openstack release for '
                      'Ubuntu release {}'.format(ubuntu_release))
        return None


def uca_openstack_releases(ubuntu_release):
    '''Return a list of OpenStack releases available to particular release
    of Ubuntu via cloud archives aka UCA.

    :param ubuntu_release: Ubuntu release codename (eg, utopic)
    :returns: OpenStack release names as list
    '''

    _map = read_yaml(MAP_FILE)
    try:
        return _map['uca-map'][ubuntu_release]
    except KeyError:
        logging.error('Could not determine OpenStack Ubuntu CloudArchive '
                      'info for Ubuntu release {}'.format(ubuntu_release))
        return None


def get_u_supported():
    '''Return a list of currently supported Ubuntu releases.
    '''
    _dist = distro_info.UbuntuDistroInfo()
    return _dist.supported()


def get_u_devel():
    '''Return the current Ubuntu development release.
    '''
    _dist = distro_info.UbuntuDistroInfo()
    return _dist.devel()


def get_u_lts():
    '''Return the current Ubuntu LTS release.
    '''
    _dist = distro_info.UbuntuDistroInfo()
    return _dist.lts()


def dict_compare(dict_old, dict_new):
    '''Compare two dictionaries, return dict of lists containing the
    original keys which have been added, removed, have changed values,
    or have the same values.
    '''
    dict_new_keys = frozenset(dict_new.keys())
    dict_old_keys = frozenset(dict_old.keys())
    intersect_keys = dict_new_keys.intersection(dict_old_keys)
    _d = {}
    _d['added'] = list(dict_new_keys - dict_old_keys)
    _d['rmvd'] = list(dict_old_keys - dict_new_keys)
    _d['diff'] = list(k for k in intersect_keys if dict_new[k] != dict_old[k])
    _d['same'] = list(k for k in intersect_keys if dict_new[k] == dict_old[k])
    _d['diff-rmvd'] = list(set(_d['diff']) - set(_d['rmvd']))
    _d['changed'] = list(set(_d['added'] + _d['diff'] + _d['rmvd']))

    if _d['added']:
        logging.debug('Added: \n{}'.format(dict_to_yaml(_d['added'])))

    if _d['rmvd']:
        logging.debug('Removed: \n{}'.format(dict_to_yaml(_d['rmvd'])))

    if _d['diff']:
        logging.debug('Diff: \n{}'.format(dict_to_yaml(_d['diff'])))

    if _d['same']:
        logging.debug('Same: \n{}'.format(dict_to_yaml(_d['same'])))
    return _d


def render(source, target, context, templates_dir=None):
    '''Render a template.

       The `source` path, if not absolute, is relative to the `templates_dir`.

       The `target` path should be absolute.

       The context should be a dict containing the values to be replaced in
       the template.

       If omitted, `templates_dir` defaults to the current working dir.
    '''

    if templates_dir is None:
        templates_dir = os.getcwd()
    loader = Environment(loader=FileSystemLoader(templates_dir))
    try:
        source = source
        template = loader.get_template(source)
    except Exception:
        logging.error('Could not load template {} from {}.'.format(
            source, templates_dir))

    safe_mkdir(templates_dir)
    with open(target, 'w') as _file:
        _file.write(template.render(context))


# Launchpad helpers
def lpl_connect(anonymous=False):
    '''Connect to launchpad, return a Launchpad object.
    '''
    if anonymous:
        logging.debug('Attempting anonymous launchpad connection...')
        launchpad = Launchpad.login_anonymously('just testing', 'production')
    else:
        logging.debug('Attempting authenticated launchpad connection...')
        launchpad = Launchpad.login_with(sys.argv[0],
                                         service_root='production',
                                         credentials_file=LPL_CRED_FN)
    logging.info('Using launchpad as: {} {}'.format(launchpad.me.name,
                                                    launchpad.me))
    return launchpad


def lpl_me(launchpad=None):
    '''Returns the current launchpad <person> object for the logged
    in user or None if logged in anonymously.  If not connected,
    connect to launchpad with default options.'''

    if not launchpad:
        launchpad = lpl_connect()

    try:
        return launchpad.me
    except Unauthorized:
        # Unauthorized is raised when logged into launchpad anonymously.
        return None


def lpl_get_mps(launchpad, lp_project_name=None,
                lp_branch_url=None, status=None):
    '''Use provided launchpad connection with project name or branch name
    to query and return a list of merge proposals.
    '''
    if lp_project_name and lp_branch_url:
        logging.debug('Usage error:  specify ONE or the other, '
                      'project name or branch name, not both.')
        return None
    else:
        if not status:
            status = ['Work in progress', 'Approved', 'Needs review',
                      'Rejected', 'Merged', 'Code failed to merge',
                      'Queued', 'Superseded']

    if lp_project_name:
        project = launchpad.projects[lp_project_name]
        proposals = project.getMergeProposals(status=status)

    if lp_branch_url:
        branch = launchpad.branches.getByUrl(url=lp_branch_url)
        proposals = branch.getMergeProposals(status=status)

    logging.debug('MPs for project: {}'.format(len(proposals)))
    return proposals


def lpl_get_mp(launchpad, mp_url, lp_branch_url):
    '''Return launchpad merge proposal object from the provided
       merge proposal url and branch url.
    '''
    branch = launchpad.branches.getByUrl(url=lp_branch_url)
    status_all = ('Work in progress/Needs review/Approved/Rejected/Merged'
                  '/Code failed to merge/Queued/Superseded'.split('/'))
    proposals = branch.getMergeProposals(status=status_all)
    logging.debug('Proposal objects found {}'.format(len(proposals)))
    for mp in proposals:
        logging.debug('Considering {} for {}'.format(mp.address,
                                                     lp_branch_url))
        if mp.web_link == mp_url:
            logging.debug('Found matching {} for {}'.format(mp.address,
                                                            lp_branch_url))
            return mp


def lpl_mp_post_comment(launchpad, mp, subject, comment, new_status=None):
    '''Post comment to merge proposal, optionally setting the status.
    '''
    mp.createComment(subject=subject, content=comment)
    if new_status:
        mp.setStatus(status=new_status)
    mp.lp_save()


def pb_it(file_name):
    '''pastebin a file_name, return pastebin url
    '''
    if os.path.isfile(file_name):
        pb_url = subprocess.Popen(['pastebinit', file_name],
                                  stdout=subprocess.PIPE).stdout
        pb_line = pb_url.readline().rstrip()
        if pb_line[:4] == 'http':
            return pb_line
        else:
            return 'pastebin not avail., cmd error'
    else:
        logging.error('file not found: {}'.format(file_name))
        return 'pastebin not avail., file error'


def ping_host(host="127.0.0.1", count=2, timeout=0.5):
    ''' Ping a host, return True if reachable.
    '''
    try:
        logging.debug('Pinging host {} with {} packets '
                      'and {}ms timeout.'.format(host, count, timeout))
        DEV_NULL = open(os.devnull, 'w')
        ret = subprocess.call(["ping", "-c", str(count),
                               "-w", str(timeout), host],
                              stdout=DEV_NULL, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        ret = e.returncode
        logging.debug('Failed to ping host {}'.format(host))

    logging.debug('Ping command returned: {}'.format(ret))
    if ret == 0:
        return True
    else:
        return False


def port_knock(host="127.0.0.1", port=22, timeout=15):
    ''' Open a TCP socket to check for a listening sevice on a host.
    '''

    # Resolve if possible
    try:
        _host = socket.gethostbyname(host)
    except socket.error as e:
        logging.debug('Unable to resolve address: '
                      '{} ({}) Trying anyway!'.format(host, e))
        _host = host

    # Include host name if applicable
    if _host == host:
        host_str = _host
    else:
        host_str = "{} ({})".format(_host, host)

    # Attempt socket connection
    try:
        knock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        knock.settimeout(timeout)
        knock.connect((_host, port))
        knock.close()
        logging.debug('Socket connect OK for host '
                      '{} on port {}.'.format(host_str, port))
        return True
    except socket.error as e:
        logging.debug('Socket connect FAIL for'
                      ' {} port {} ({})'.format(host_str, port, e))
        return False


def ssh_command_check(host='127.0.0.1', username=None,
                      password=None, cmd='uname', expected=['Linux\n'],
                      pkey=None, timeout=10, mpq=None):
    ''' Run a command via SSH, return true if successful.
    '''
    logging.debug('SSH connect check to host: {} with command '
                  '{}'.format(host, cmd))

    try:
        logging.getLogger("paramiko").setLevel(logging.WARNING)
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        if pkey:
            pkey_rsa = paramiko.RSAKey.from_private_key_file(pkey)
            logging.debug('Using SSH key: {}'.format(pkey))
        else:
            pkey_rsa = None

        ssh.connect(host, username=username, password=password,
                    pkey=pkey_rsa, timeout=timeout)
        stdin, stdout, stderr = ssh.exec_command(cmd)
        stdout_lines = stdout.readlines()
        stderr_lines = stderr.readlines()
        logging.debug('SSH user: {}'.format(username))
        if password:
            logging.debug('SSH password specified.')
        logging.debug('SSH expected: {}'.format(expected))
        logging.debug('SSH stdout: {}'.format(stdout_lines))
        logging.debug('SSH stderr: {}'.format(stderr_lines))
    except paramiko.ssh_exception.SSHException as e:
        logging.error('SSH session connect FAILED for host '
                      '{} ({})'.format(host, e))
        if mpq:
            mpq.put(False)
        return False
    except socket.error as e:
        logging.error('SSH socket connect FAIL for'
                      ' {} ({})'.format(host, e))
        if mpq:
            mpq.put(False)
        return False

    # Inspect for debug output
    if stdout_lines:
        if stdout_lines == expected:
            logging.debug('SSH command output matches expected output.')
        else:
            logging.debug('SSH command output does not match expected output.')
    else:
        logging.debug('SSH command had no stdout content.')

    if stderr_lines:
        logging.debug('SSH command had error output.')

    # Final validation
    if stdout_lines and stdout_lines == expected and not stderr_lines:
        if mpq:
            mpq.put(True)
        return True
    else:
        if mpq:
            mpq.put(False)
        return False


def func_timeout(timeout=2, func_call=None, args=(), kwargs={}):
    ''' Call a function and impose a timeout (s).  Returns True if the process
        completed successfully within the specified timeout period (seconds).

        :param timeout: Seconds to allow function to run before terminating
        :param func_call: Function to call
        :param args: Tuple of function arguments
        :param kwargs: Dictionary of function keyword:value pairs
        :return: True if the process completed within the timeout (sec).
    '''

    # Call function as a process
    _queue = multiprocessing.Queue()
    kwargs['mpq'] = _queue
    _func = multiprocessing.Process(target=func_call,
                                    args=args,
                                    kwargs=kwargs)
    _func.start()

    # Terminate after timeout period
    _func.join(timeout)
    if _func.is_alive():
        logging.debug('func_timeout, terminated {} for exceeding '
                      '{}s.'.format(func_call, timeout))
        _func.terminate()
        return False

    # Check return code
    logging.debug(_func.exitcode)
    if _func.exitcode != 0:
        logging.debug('func_timeout, failing for non-zero exit code: '
                      '{}'.format(_func.exitcode))
        return False
    else:
        _queue_return = _queue.get()
        logging.debug('func_timeout OK.  Queue: {}'.format(_queue_return))
        return _queue_return


# branch scanner helpers
def dict_to_yaml(dict_data):
    return yaml.dump(dict_data, default_flow_style=False)


def create_properties_files(contexts, source, out_file,
                            templates_dir='populate'):
    '''Create properties files for targets defined in contexts dict.'''
    for idx, context in enumerate(contexts):
        if 'mojo_spec' in context.keys() and 'metal' in context['mojo_spec']:
            _file = 'baremetal-{}'.format(out_file)
        else:
            _file = out_file

        target = os.path.join(WORKSPACE, _file.format(uuid.uuid4()))
        logging.debug('Rendering properties file: {}'.format(target))
        render(source, target, context, templates_dir)
    logging.info('Created {} properties files.'.format(len(contexts)))


def is_charm_branch(branch_str):
    '''Return true if the branch string looks like a charm.'''
    if '/charms/' in branch_str or 'lp:charms/' in branch_str:
        logging.debug(' + Looks like a charm: {}'.format(branch_str))
        return True
    else:
        logging.debug(' - Does not look like a charm: {}'.format(branch_str))
        return False


def is_next_branch(branch_str):
    '''Return True if the branch string looks like a next charm.'''
    if branch_str[-5:] == "/next" in branch_str:
        logging.debug(' + Looks like a next branch: {}'.format(branch_str))
        return True
    else:
        logging.debug(' - Does not look like a next branch: '
                      '{}'.format(branch_str))
        return False


def is_trunk_branch(branch_str):
    '''Return True if the branch string looks like a next charm.'''
    if branch_str[-6:] == "/trunk" in branch_str:
        logging.debug(' + Looks like a trunk branch: {}'.format(branch_str))
        return True
    else:
        logging.debug(' - Does not look like a trunk branch: '
                      '{}'.format(branch_str))
        return False


def get_tags(branch_str):
    '''Return a list of hash tags in a string.'''
    if '#' in branch_str:
        return branch_str.split('#')[1:]
    else:
        return None


def strip_tags(branch_str):
    '''Return a string without hash tags.'''
    if '#' in branch_str:
        return branch_str.split('#')[0]
    else:
        return branch_str


def dict_strip_tags(dict_data):
    '''Remove tags from keys in a dictionary.'''
    _d = {}
    for k, v in dict_data:
        _d[strip_tags(k)] = v
    return _d


def list_strip_tags(list_data):
    '''Remove tags from values in a list of strings.'''
    _list = []
    for v in list_data:
        _list.append(strip_tags(v))
    return _list


def is_tagged(my_string, tagged, tags):
    '''Check if a string contains any of a list of tags, based on
    a dict of tags to branch strings.
    '''
    for tag in tags:
        if my_string in tagged[tag]:
            logging.info('Tag: {} on {}'.format(tag, my_string))
            return True
    return False


def get_tagged_branches(conf_map):
    '''Get tagged branches from config map data.  Return a dict of
    tag keys and branch values.
    '''
    # Available tags
    tagged = {
        'NOAMULET': [],
        'NODEPLOY': [],
        'NOLINT': [],
        'NOUNIT': [],
    }

    for branch_str in conf_map['trigger-branches']:
        tags = get_tags(branch_str)
        if not tags:
            continue
        for tag in tags:
            if tag not in tagged.keys():
                raise ValueError('Invalid tag: {}.  Valid tags '
                                 'are: {}'.format(tag, tagged.keys()))
            tagged[tag].append(strip_tags(branch_str))

    logging.debug('Tagged: \n{}'.format(dict_to_yaml(tagged)))
    return tagged


def get_next_base_name(branch_str):
    '''Return charm name, given a next branch string.'''
    return branch_str.split('/')[-2]


def get_stable_base_name(branch_str):
    '''Return charm name, given a non-next branch string.'''
    return strip_chars('-'.join(branch_str.split('/')[-1:]))


def get_trunk_base_name(branch_str):
    '''Return charm name, given a next branch string.'''
    return branch_str.split('/')[-2]


def create_properties_dict(changed, tagged={},
                           blacktags=[], charms_only=False):
    '''Generate properties data which can be passed to a make lint-
    properties, unit-properties, amulet-properties files, routines, etc.
    for consumption by Jenkins parameterized jobs.  Accepts dict of tagged
    branches and a list of blacklistedm, to skip matches.  Optionally
    create properties only for charm branches.
    '''
    base_name = None
    display_name = None
    contexts = []
    changed.sort()
    for branch_str in changed:

        # Skip branches tagged with blacktags
        if is_tagged(branch_str, tagged, blacktags):
            continue

        is_charm = is_charm_branch(branch_str)
        is_next = is_next_branch(branch_str)
        is_trunk = is_trunk_branch(branch_str)

        if is_charm and is_next:
            # lp:~openstack-charmers/charms/trusty/glance/next
            base_name = get_next_base_name(branch_str)
            display_name = '{}-next'.format(base_name)
        elif is_charm and is_trunk:
            # lp:~somebody/charms/trusty/some-charm/trunk
            base_name = get_trunk_base_name(branch_str)
            display_name = base_name
            display_name = '{}-trunk'.format(base_name)
        elif is_charm and not is_next and not is_trunk:
            # lp:charms/trusty/glance
            base_name = get_stable_base_name(branch_str)
            display_name = base_name
        else:
            base_name = branch_str
            display_name = branch_str

        if (is_charm and charms_only) or not charms_only:
            contexts.append({
                'branch': branch_str,
                'base_name': base_name,
                'display_name': display_name
            })
    return contexts


def create_deploy_properties_dict(deploy_group):
    '''Generate template context data for deploy targets in a provided dict.
    Return a dictionary which can be passed to a make deploy-properties
    files routine.
    '''
    rls_supp = get_u_supported()
    contexts = []
    for charm_set, targets in deploy_group.iteritems():
        if not targets:
            continue

        targets.sort(reverse=True)
        for target in targets:
            target_split = target.split('-')
            u_rls = target_split[0]
            os_rls = target_split[1]

            if len(target_split) == 3:
                inst_src = target_split[2]

            elif len(target_split) == 2:
                inst_src = 'default'
            else:
                raise ValueError('Invalid target name: {}'.format(target))

            if u_rls not in rls_supp:
                logging.warn('Detected an unsupported '
                             'release: {}'.format(u_rls))

            contexts.append({
                'charm_set': charm_set,
                'u_rls': u_rls,
                'os_rls': os_rls,
                'inst_src': inst_src
            })
    return contexts


def get_state_file_name(conf_map, key='state-file'):
    '''Get branch scanner state file name based on config data.'''
    safe_mkdir(conf_map['misc-conf']['state-dir'])
    state_fn = os.path.join(conf_map['misc-conf']['state-dir'],
                            conf_map['misc-conf'][key])

    logging.info('State file: {}'.format(state_fn))
    return state_fn


def read_state_file(state_fn):
    '''Read state data from state file.'''
    logging.debug('Reading state data from: {}'.format(state_fn))
    if os.path.isfile(state_fn):
        state_prev = read_yaml(state_fn)
    else:
        state_prev = {}
    return state_prev


def write_state_file(state_data, state_fn):
    '''Commit state data to the state file.'''
    logging.debug('Writing state data to: {}'.format(state_fn))
    write_yaml(state_data, state_fn)


def write_changed_file(changed_data, changed_fn):
    '''Write a flat file containing new or changed branch strings.'''
    logging.debug('Writing added/changed info to: {}'.format(changed_fn))
    with open(changed_fn, 'w') as w_file:
        w_file.write('\n'.join(changed_data))


def fake_branch_states(branch_list):
    '''Construct a dict with branch strings as keys, and epoch time as
    values.  Useful for forcing all to appear to be changed.
    '''
    state_fake = {}
    for branch_str in list_strip_tags(branch_list):
        logging.info('Forcing simulated change on:  {}'.format(branch_str))
        state_fake[branch_str] = time.time()
    return state_fake


def query_branch_states(branch_list, max_tries=10):
    '''Get current state of a list of launchpad branches, return a dict.'''
    load_plugins()
    state_now = {}
    for branch_str in list_strip_tags(branch_list):
        logging.info('Querying revision info on:  {}'.format(branch_str))
        ret_ok, tries = False, 0

        # Check branches and handle bazaar http retries
        while not ret_ok and tries <= max_tries:
            try:
                this_br = branch.Branch.open(branch_str)
                ret_ok = True
            except bzrlib.errors.InvalidHttpResponse as exception:
                logging.debug('HTTP error checking {} on attempt {} of '
                              '{}'.format(branch_str, tries, max_tries))
                logging.error('{} {}'.format(exception.path, exception.msg))
                time.sleep(3)
            tries += 1

        if not ret_ok:
            raise ValueError('Query failed on:  {}'.format(branch_str))

        state_now[branch_str] = list(this_br.last_revision_info())

    return state_now


def is_deploy_needed(changed, tagged={}, blacktags=[]):
    '''Analyze strings of changed branches to determine if any next
    or stable charms have changed, while excluding those blacklisted
    with a tag.'''

    if not changed:
        return False

    deploy_stable, deploy_next = False, False
    for branch_str in changed:

        # Skip branches tagged with blacktags
        if is_tagged(branch_str, tagged, blacktags):
            continue

        is_charm = is_charm_branch(branch_str)
        is_next = is_next_branch(branch_str)

        if is_charm and is_next:
            deploy_next = True
        elif is_charm and not is_next:
            deploy_stable = True

#   TODO - trigger next and stable independently
#   For now, a change in either with trigger both.
#    logging.info('Deploy stable: {}'.format(deploy_stable))
#    logging.info('Deploy next: {}'.format(deploy_next))
#    return deploy_next, deploy_stable

    logging.info('Deploy: {}'.format(deploy_stable or deploy_next))

    if deploy_next or deploy_stable:
        return True
    else:
        return False


def define_scanner_command_options(USAGE):
    '''Define command line options.'''
    parser = optparse.OptionParser(USAGE)
    parser.add_option("-d", "--debug",
                      help="Enable debug logging",
                      dest="debug", action="store_true", default=False)
    parser.add_option('--XA', '--no-amulet-trigger',
                      help='Do not create amulet job properties files.',
                      dest='amulet', action='store_false', default=True)
    parser.add_option('--XD', '--no-deploy-trigger',
                      help='Do not create deploy job properties files.',
                      dest='deploy', action='store_false', default=True)
    parser.add_option('--XL', '--no-lint-trigger',
                      help='Do not create lint job properties files.',
                      dest='lint', action='store_false', default=True)
    parser.add_option('--XU', '--no-unit-trigger',
                      help='Do not create unit test job properties files.',
                      dest='unit', action='store_false', default=True)
    parser.add_option('-F', '--force-changed',
                      help='Do not check branches, act like ALL were changed '
                           'regardless of actual state.  Default False.',
                      dest='force', action='store_true', default=False)
    parser.add_option('-X', '--no-scan',
                      help='Do not check branches, act like NONE were changed '
                           'regardless of actual state.  Default False.',
                      dest='no_scan', action='store_true', default=False)
    parser.add_option('-N', '--no-state',
                      help='Do not update state file.  Has no effect if '
                           'used in combination with dry run.',
                      dest='no_state', action='store_true', default=False)
    parser.add_option('-0', '--dry-run',
                      help='Do not create properties files, and do not '
                           'update state file(s).',
                      dest='dry', action='store_true', default=False)

    return parser.parse_args()


def handle_scanner_command_options(opts, args):
    '''Handle command options and inform user.'''
    if opts.debug:
        logging.basicConfig(level=logging.DEBUG)
        logging.info('Logging level set to DEBUG!')
        logging.debug('parse opts: \n{}'.format(dict_to_yaml(vars(opts))))
        logging.debug('parse args: {}'.format(args))
    else:
        logging.basicConfig(level=logging.INFO)

    logging.info('Workspace dir: {}'.format(WORKSPACE))

    if opts.force:
        logging.info(' - Branch scanning disabled.  Will force state to '
                     'CHANGED for all monitored branches!')
    else:
        logging.info(' + Branch scanning enabled.')

    if opts.amulet:
        logging.info(' + Amulet trigger creation enabled.')
    else:
        logging.info(' - Amulet trigger creation disabled.')

    if opts.unit:
        logging.info(' + Unit test trigger creation enabled.')
    else:
        logging.info(' - Unit test trigger creation disabled.')

    if opts.lint:
        logging.info(' + Lint trigger creation enabled.')
    else:
        logging.info(' - Lint trigger creation disabled.')

    if opts.deploy:
        logging.info(' + Deploy trigger creation enabled.')
    else:
        logging.info(' - Deploy trigger creation disabled.')

    if opts.no_state:
        logging.info(' - No-Update enabled, state files will not be updated.')

    if opts.dry:
        logging.info(' - Dry run enabled, properties files and state files '
                     'will not be created/updated.')

    if opts.no_scan and opts.force:
        logging.warn('Specifying Force and No Scan makes no sense.  YMMV')
