#!/usr/bin/env python
'''
Creates properties files based on mojo trigger groups defined in
mappings.yaml, for consumption by jenkins in parameterized builds.
'''

import optparse
import logging
import yaml
import os
import sys
# import traceback
import tempfile
from common import osci_utils
# from copy import deepcopy
from time import sleep
from urllib2 import urlopen
import hashlib

USAGE = '''Usage: %prog [options]

Creates properties files based on URL trigger groups defined in
mappings.yaml, for consumption by jenkins in parameterized builds.

For example:

    %prog -d -l
  Lists all available URL trigger groups, with debug output.

    %prog -g "jprop"
  Generates properties files for all iterations defined in the
  specified group name from mappings.yaml.

    %prog -0d -g "your_group"
  Generates properties data dictionaries for all iterations defined
  in the specified group name from mappings.yaml, with debug output,
  but does not create properties files.  Useful for troubleshooting.
'''

try:
    WORKSPACE = os.environ['WORKSPACE']
except KeyError:
    WORKSPACE = tempfile.mkdtemp()


# distill and move to osci_common!   then update other tools to use it.
def create_properties_dict(changed, state, charms_only=False):
    '''Generate context dict for creation of lint-properties,
    unit-properties, and amulet-properties files.
    '''
    contexts = []
    changed.sort()
    for mp_src in changed:
        _is_charm = False
        mp_tgt_br = state[mp_src]['scanned-branch']
        mp_src_br = state[mp_src]['src']['display']
        mp_reg = state[mp_src]['mp-registrant']
        mp_num = osci_utils.strip_chars(mp_src.split('@')[0])

        if '/charms/' in mp_tgt_br or 'lp:charms/' in mp_tgt_br:
            logging.debug(' + Looks like a charm: {}'.format(mp_tgt_br))
            _is_charm = True
        else:
            logging.debug(' - Does not look like '
                          'a charm: {}'.format(mp_tgt_br))
        if mp_tgt_br[-5:] == "/next" and '/charms/' in mp_tgt_br:
            _base_name = mp_tgt_br.split('/')[-2]
            _display_name = '{}-next for {} {}'.format(_base_name,
                                                       mp_reg,
                                                       mp_num)
        else:
            _base_name = osci_utils.strip_chars('-'.join(
                                                mp_tgt_br.split('/')[-1:]))
            _display_name = '{} for {} {}'.format(_base_name,
                                                  mp_reg,
                                                  mp_num)
        if (_is_charm and charms_only) or not charms_only:
            contexts.append({
                'branch': mp_src_br,
                'base_name': _base_name,
                'display_name': _display_name,
                'mp_web': state[mp_src]['mp-web'],
                'mp_src_br': mp_src_br,
                'mp_tgt_br': mp_tgt_br
            })
    return contexts


# distill and move to osci_common!   then update other tools to use it.
def create_properties_files(contexts, source, out_file):
    '''Create properties files for targets defined in contexts dict
    '''
    templates_dir = 'populate'

    for idx, context in enumerate(contexts):
        target = os.path.join(WORKSPACE, out_file.format(idx))
        logging.debug('Rendering properties file: {}'.format(target))
        osci_utils.render(source, target, context, templates_dir)
        sleep(0.05)


def process_url_trigger_group(conf_map, opts):
    t_group = conf_map['url-trigger-groups'][opts.group]

    # read state file if it exists
    osci_utils.safe_mkdir(conf_map['misc-conf']['state-dir'])
    state_fn = os.path.join(conf_map['misc-conf']['state-dir'],
                            t_group['state-file'])

    if os.path.isfile(state_fn):
        logging.debug('Reading URL state from file {}.'.format(state_fn))
        url_state_prev = osci_utils.read_yaml(state_fn)
    else:
        logging.debug('No URL state file found, starting clean.')
        url_state_prev = {}

    # get current url state
    url_state = {}
    url_state[opts.group] = {}
    for this_url in t_group['urls']:
        # this_hash = osci_utils.rnd_str(length=32)
        this_hash = url_contents_hash(this_url)
        url_state[opts.group][this_url] = this_hash

    # Compare current URL group state with last state
    comps = osci_utils.dict_compare(url_state_prev,
                                    url_state)
    changed = list(set(comps['added'] + comps['diff']) - set(comps['rmvd']))
    url_changed_fn = os.path.join(conf_map['misc-conf']['state-dir'],
                                  '{}.changed'.format(state_fn))

    if len(comps['added']) > 0:
        logging.info('added: \n{}'.format(yaml.dump(comps['added'],
                                          default_flow_style=False)))
    if len(comps['diff']) > 0:
        logging.info('changed: \n{}'.format(yaml.dump(comps['diff'],
                                            default_flow_style=False)))
    if len(comps['same']) > 0:
        logging.debug('unchanged: \n{}'.format(yaml.dump(comps['same'],
                                               default_flow_style=False)))
    if len(comps['rmvd']) > 0:
        logging.info('removed: \n{}'.format(yaml.dump(comps['rmvd'],
                                            default_flow_style=False)))
    if len(changed) > 0:
        logging.debug('added or changed: \n{}'.format(
            yaml.dump(changed, default_flow_style=False)))

    if (len(changed) > 0 or len(comps['rmvd']) > 0):
        if not opts.dry:
            logging.debug('Writing URL state file.')
            # Write current branch state to file
            osci_utils.write_yaml(url_state, state_fn)
            # Also write flat file containing new or changed URL groups
            if len(changed) > 0:
                with open(url_changed_fn, 'w') as w_file:
                    w_file.write('\n'.join(changed))
        else:
            logging.debug('Not writing URL state file due to options.')
    else:
        # no changes detected
        logging.info('No URL changes detected.')
        if os.path.isfile(url_changed_fn):
            os.remove(url_changed_fn)
        changed, url_state = None, None

    # trigger something

    logging.debug(yaml.dump(url_state, default_flow_style=False))


def url_contents_hash(url=None):
    '''Fetch html content from a URL, return its hash string.
    '''
    _buffer = urlopen(url)
    _html = _buffer.read()
    _hash = hashlib.sha256(_html).hexdigest()
    return _hash


def main():
    '''Define and handle command line parameters
    '''
    # Define command line options
    parser = optparse.OptionParser(USAGE)
    parser.add_option("-d", "--debug",
                      help="Enable debug logging",
                      dest="debug", action="store_true", default=False)
    parser.add_option("-g", "--group",
                      help="URL trigger group as defined in mappings.yaml",
                      action="store", type="string", dest="group")
    parser.add_option("-l", "--list",
                      help="List all available URL trigger groups.  Can not "
                           "be used with -g or -0.",
                      dest="list", action="store_true", default=False)
    parser.add_option('-0', '--dry-run',
                      help='Do not create properties files.',
                      dest='dry', action='store_true', default=False)

    conf_map = osci_utils.read_yaml(osci_utils.MAP_FILE)

    params = parser.parse_args()
    (opts, args) = params

    # Validate options
    if not opts.group and not opts.list:
        parser.print_help()
        sys.exit(1)

    # Handle parameters, inform user
    if opts.debug:
        logging.basicConfig(level=logging.DEBUG)
        logging.info('Logging level set to DEBUG!')
        logging.debug('parse opts: \n{}'.format(
            yaml.dump(vars(opts), default_flow_style=False)))
        logging.debug('parse args: {}'.format(args))
    else:
        logging.basicConfig(level=logging.INFO)

    logging.info('Workspace dir: {}'.format(WORKSPACE))

    # list url trigger groups
    if opts.list:
        logging.info('List URL Groups...')
        for group, contents in conf_map['url-trigger-groups'].iteritems():
            logging.info('URL group: {} (with {} '
                         'urls)'.format(group, len(contents['urls'])))
        logging.debug('URL trigger groups:\n{}'.format(
            yaml.dump(conf_map['url-trigger-groups'],
                      default_flow_style=False)))

    if opts.dry:
        logging.info('Dry run!')
        pass

    if opts.group:
        if opts.group in conf_map['url-trigger-groups']:
            # validated, go!
            process_url_trigger_group(conf_map, opts)
        else:
            raise ValueError('URL group not found: {}\n'
                             'Available groups: {}'.format(
                                 opts.group,
                                 conf_map['url-trigger-groups'].keys()))


#    if changed and opts.unit:
#        unit_targets = create_properties_dict(changed, state)
#        logging.debug('unit targets: \n{}'.format(
#            yaml.dump(unit_targets, default_flow_style=False)))
#
#    if changed and opts.unit and not opts.dry:
#        logging.info('Creating unit-properties files')
#        create_properties_files(unit_targets,
#                                source='unit-properties.tmpl',
#                                out_file='unit-properties.{}')


if __name__ == '__main__':
    print('Ubuntu OSCI URL Group Scanner')
    main()


'''
url-trigger-groups:
  jprop:
    # juju proposed ppa contents watcher
    state-file: 'jprop-state'
    urls:
    - http://ppa.launchpad.net/juju/proposed/ubuntu/pool/main/j/juju-core/
    # and temp urls for increased testing of the tester:
    - https://launchpad.net/~1chb1n/+karma
    - https://code.launchpad.net/~uosci/ubuntu-openstack-ci/trunk
    - http://irclogs.ubuntu.com/2015/02/25/%23juju-dev.html
    - http://irclogs.ubuntu.com/2015/02/25/%23juju.html
    - http://irclogs.ubuntu.com/2015/02/25/%23ubuntu-server.html
'''
