#!/usr/bin/env python
'''
Find text string matches in a file, replace with text,
write to a different file.
'''

import optparse
import logging
import yaml
import os
import sys

USAGE = '''Usage: %prog [options]

Default behavior:
  Find string matches in a file, replace with text, write to a different file,
  with optional debug output.

For example:

    %prog -yd -i README -o README2 -f "some text" -r "replace with this"

  Replaces matching text in the README file, saves to README2, overwriting
  if README2 already exists, with debug output.
'''


def main():
    '''Define and handle command line parameters
    '''
    # Define command line options
    parser = optparse.OptionParser(USAGE)
    parser.add_option("-d", "--debug",
                      help="Enable debug logging",
                      dest="debug", action="store_true", default=False)
    parser.add_option('-0', '--dry-run',
                      help='Do not create, write or modify any files',
                      dest='dry', action='store_true', default=False)
    parser.add_option('-y', '--yes-overwrite',
                      help='Overwrite the output file',
                      dest='overwrite', action='store_true', default=False)
    parser.add_option("-i", "--in-file",
                      help="Deployer bundle input (source) file",
                      action="store", type="string", dest="in_file")
    parser.add_option("-o", "--out-file",
                      help="Deployer bundle output (destination) file",
                      action="store", type="string", dest="out_file")
    parser.add_option("-f", "--find",
                      help="String to find",
                      action="store", type="string", dest="find")
    parser.add_option("-r", "--replace",
                      help="String with which to replace the found string(s)",
                      action="store", type="string", dest="replace")

    params = parser.parse_args()
    (opts, args) = params

    # Handle parameters, inform user
    if opts.debug:
        logging.basicConfig(level=logging.DEBUG)
        logging.info('Logging level set to DEBUG!')
        logging.debug('parse opts: \n{}'.format(
            yaml.dump(vars(opts), default_flow_style=False)))
        logging.debug('arg count: {}'.format(len(args)))
        logging.debug('parse args: {}'.format(args))
    else:
        logging.basicConfig(level=logging.INFO)

    # Validate options
    if (not opts.find or not opts.replace or not
            opts.in_file or not opts.out_file):
        parser.print_help()
        sys.exit(1)

    # Validate and process files
    if opts.out_file:
        if os.path.isfile(opts.out_file) and opts.overwrite:
            logging.warning('Output file exists and will be '
                            'overwritten: {}'.format(opts.out_file))
        elif os.path.isfile(opts.out_file) and not opts.overwrite:
            logging.error('Output file exists and will NOT '
                          'be overwritten: {}'.format(opts.out_file))
            sys.exit(1)

    if opts.in_file:
        if os.path.isfile(opts.in_file):
            # Read the infile
            logging.info('reading input file: {}'.format(opts.in_file))
            with open(opts.in_file) as in_file:
                in_lines = in_file.readlines()
                out_lines, found = [], 0
                # Replace matching text
                for line in in_lines:
                    if opts.find in line:
                        found += 1
                        mod_line = line.replace(opts.find, opts.replace)
                        logging.debug('changing line: \n{}'
                                      '  to:\n{}'.format(line, mod_line))
                        out_lines.append(mod_line)
                    else:
                        out_lines.append(line)
            if found > 0:
                # Write the outfile
                with open(opts.out_file, 'w') as out_file:
                    logging.info('writing out file: {}'.format(opts.out_file))
                    out_file.writelines(out_lines)
                logging.info('replaced {} lines'.format(found))
            else:
                logging.error('No matches found, output file not written.')
                sys.exit(1)
        else:
            logging.error('Input file not found: {}'.format(opts.in_file))
            sys.exit(1)


if __name__ == '__main__':
    main()
