#!/usr/bin/python
'''
  Pull logs from all juju units, store in WORKSPACE if env var is set,
  else store in local temp directory.
'''
import os
import logging
import sys
sys.path.insert(1, os.path.join(sys.path[0], '..'))
from common import osci_utils as u  # noqa

# Set default logging level
logging.basicConfig(level=logging.DEBUG)


def run_cmds(cmds, fatal=False, dry_run=False, suppress=True):
    ''' Execute a list of commands.  Use os.system instead of subprocess
        due to piping and redirection.
    '''

    if not cmds:
        return False

    all_passed = True
    for cmd in cmds:
        if suppress:
            cmd = cmd + ' >/dev/null 2>&1'

        # Run command
        logging.info('running cmd: {}'.format(cmd))
        if not dry_run:
            ret = os.system(cmd)

        # Handle exit code and fatal flag
        if fatal and ret:
            logging.warn('Aborting on failed command.')
            sys.exit(1)
        elif not fatal and ret:
            all_passed = False
            logging.warn('Ignoring failed command.')

    if all_passed:
        return True
    else:
        return False


def main():
    '''You are here.  Go!
    '''

    logging.info('Ubuntu OSCI Juju Log Puller')
    logging.info('Workspace dir: {}'.format(u.WORKSPACE))
#    pkey = os.path.expanduser('~/.juju/ssh/juju_id_rsa')
    all_collected = True

    # Determine output directory
    log_dir = os.path.join(u.WORKSPACE, 'logs')

    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    logging.info('Logs dir: {}'.format(log_dir))

    # Get juju status data
    j_stat = u.juju_stat()

    # Basic jstat data validation
    if not j_stat:
        logging.info('No action taken.  Exiting cleanly, '
                     'presuming not bootstrapped.')
        sys.exit(0)
    elif j_stat and 'machines' not in j_stat.keys() or \
            'services' not in j_stat.keys():
        logging.error('No juju status data to inspect, although '
                      'an environment appears to be bootstrapped.')
        sys.exit(1)

    # Add bootstrap machine to units dict
    found_bootstrap_node = False

    for machine, mach_data in j_stat['machines'].iteritems():
        if 'state-server-member-status' in mach_data:
            if mach_data['state-server-member-status'] == 'has-vote':
                found_bootstrap_node = True
                bs_mach = machine
                logging.info('bootstrap machine: {}'.format(machine))
                j_stat['services']['bootstrap'] = {
                    'units': {
                        bs_mach: {
                            'agent-state': mach_data['agent-state'],
                            'machine': bs_mach
                        }
                    }
                }

    if not found_bootstrap_node:
        logging.error('Could not identify bootstrap node!')

    # Build commands list from juju status data
    cmds = []
    for charm, services in j_stat['services'].iteritems():
        if 'units' not in services.keys():
            logging.info('No unit for {}'.format(charm))
            continue

        for unit, unit_data in services['units'].iteritems():
            clean_name = unit.replace('/', '-')
            machine_num = unit_data['machine']

            # Get nova console logs if instance is nova
            if machine_num in j_stat['machines'] and \
                    'instance-id' in j_stat['machines'][machine_num]:
                instance_id = j_stat['machines'][machine_num]['instance-id']
                logging.info('Instance id for {}: {}'.format(machine_num,
                                                             instance_id))
                if 'pending' in instance_id:
                    all_collected = False
                    logging.warn('Skipping {} ({})'.format(clean_name,
                                                           instance_id))
                    continue

                if 'MAAS' not in instance_id:
                    cmds.append('nova console-log {} | bzip2 -9z | tee {}/{}'
                                '-nova-console-{}.bz2'.format(
                                    instance_id,
                                    log_dir,
                                    clean_name,
                                    unit_data['machine']))

# Temporarily disabled
# Need to refactor for later juju versions, new agent states.
#            if unit_data['agent-state'] != 'started':
#                logging.warn('Agent down for {}, '
#                             'not pulling its logs.'.format(unit))
#                all_collected = False
#                continue

# Check if ssh works before attempting to juju ssh
# !Buggy! - paramiko old version
#            if 'public-address' in unit_data.keys():
                # SSH command check on unit's host address
                # Don't try to "juju ssh" to a unit that is defunct.
#                host = unit_data['public-address']
#                logging.info('Host, key:  {}, {}'.format(host, pkey))
#                ret = u.func_timeout(
#                    timeout=10,
#                    func_call=u.ssh_command_check,
#                    kwargs={'host': host,
#                            'username': 'ubuntu',
#                            'pkey': pkey})
#
#                if not ret:
#                    logging.warn('SSH command check failed for unit {} ({}), '
#                                 'not pulling its logs.'.format(unit, host))
#                    all_collected = False
#                    continue

            # NOTE:  no-file-changed prevents non-zero exit when a file
            # changes as it is being tarred.  Just plow on.
            cmds.append('juju ssh {} sudo tar -cjf '
                        '/home/ubuntu/{}-var-log.tar.bz2 /var/log '
                        '--warning=no-file-changed'.format(
                            unit,
                            clean_name))

            cmds.append('juju ssh {} sudo tar -cjf '
                        '/home/ubuntu/{}-etc.tar.bz2 /etc '
                        '--warning=no-file-changed '
                        '--exclude="/etc/X11" --exclude="/etc/ssl" '
                        '--exclude="/etc/ssh" --exclude="shadow"'.format(
                            unit,
                            clean_name))

            # May not always exist (ceph stores confs here)
            cmds.append('juju ssh {} "[[ -d /var/lib/charm ]] && '
                        'sudo tar -cjf /home/ubuntu/{}-var-lib-charm.tar.bz2 '
                        '/var/lib/charm ||:"'.format(unit,
                                                     clean_name))

            cmds.append('juju ssh {} "dpkg -l | bzip2 -9z > '
                        '/home/ubuntu/{}-dpkg-list.bz2"'.format(unit,
                                                                clean_name))

            cmds.append('juju ssh {} "df -h | bzip2 -9z > '
                        '/home/ubuntu/{}-df.bz2"'.format(unit,
                                                         clean_name))

            cmds.append('juju ssh {} "ps aux | bzip2 -9z > '
                        '/home/ubuntu/{}-processes.bz2"'.format(unit,
                                                                clean_name))

            cmds.append('juju ssh {} "sudo netstat -taupn | '
                        'grep LISTEN | bzip2 -9z > '
                        '/home/ubuntu/{}-listening.bz2"'.format(unit,
                                                                clean_name))

            cmds.append('juju scp {}:/home/ubuntu/*.bz2'
                        ' {}'.format(unit, log_dir))

    # Run!
    failed = False
    ret = run_cmds(cmds, dry_run=False)

    # Final validation
    if not ret:
        logging.error('One or more commands failed.')
        failed = True

    if not all_collected:
        logging.error('One or more units were not reachable.')
        failed = True

    if not failed:
        logging.info('Unit log collection appeared to go well.  Good day.')
    else:
        sys.exit(1)


if __name__ == '__main__':
    main()
