#!/usr/bin/env python
'''
  juju status health check (not to be confused with juju health status)

  A quick health check based on juju agent states reported in juju stat, plus
  SSH and juju agent TCP port connection checks, and ping connectivity checks.
'''
import logging
import optparse
import os
import subprocess
import sys
import yaml
sys.path.insert(1, os.path.join(sys.path[0], '..'))
from common import osci_utils  # noqa


USAGE = '''Usage: %prog [options]

  juju status health check (not to be confused with juju health status)

  A quick health check based on juju agent states reported in juju stat, plus
  SSH and juju agent TCP port connection checks, and ping connectivity checks.
'''


def jshc_ssh_command_check(host='127.0.0.1'):
    ''' A locally reusable chunk to check SSH command success on a remote host.
    '''

    pkey = os.path.expanduser('~/.juju/ssh/juju_id_rsa')

    ret = osci_utils.func_timeout(timeout=10,
                                  func_call=osci_utils.ssh_command_check,
                                  kwargs={'host': host,
                                          'username': 'ubuntu',
                                          'pkey': pkey})

    if ret:
        logging.info('SSH command check OK for {}.'.format(host))
        return True, ''
    else:
        logging.warn('SSH command check FAIL for'
                     ' {}'.format(host))
    return False, 'host:  {}'.format(host)


def jshc_tcp_socket_check(host="127.0.0.1", port=22, timeout=15):
    ''' A locally reusable chunk to check TCP socket connectivity
        and give feedback.
    '''
    if osci_utils.port_knock(host=host, port=port, timeout=timeout):
        logging.info('Socket connect OK for {} port {}.'.format(host, port))
        return True, ''
    else:
        logging.warn('Socket connect FAIL for'
                     ' {} port {}'.format(host, port))
    return False, 'host:  {}'.format(host)


def jshc_ping_check(host, check_hist):
    ''' A locally reusable chunk to check host pingability and give feedback.
    '''
    check_str = 'ping: {}'.format(host)
    if check_str in check_hist:
        logging.info('Already checked  {}'.format(check_str))
        return None, None, check_str
    else:
        if osci_utils.ping_host(host=host):
            logging.info('Ping OK for:  {}'.format(host))
            return True, None, check_str
        else:
            logging.warn('Ping FAILED for:  {}'.format(host))
            err_str = 'host:  {}'.format(host)
        return False, err_str, check_str


def jshc_validate():
    ''' Misc validation.
    '''
    try:
        _jstat = osci_utils.juju_stat()
    except subprocess.CalledProcessError:
        logging.error('Could not get juju status.')
        sys.exit(1)

    if 'services' not in _jstat:
        sys.exit(2)


def juju_status_health_check(ignore_bootstrap_node=False):
    ''' Check juju agent, instance, machine status and connectivity.
    '''

    jshc_validate()
    j_stat = osci_utils.juju_stat()

    sick = []
    check_hist = set()
    found_bootstrap_node = False

    # per-service unit state check loop
    logging.debug('Checking agents per service.')
    for charm, services in j_stat['services'].iteritems():
        logging.info('=== Checking service: {} ==='.format(charm))

        # check for existing units
        if 'units' not in services.keys():
            logging.info('No unit(s) for {}'.format(charm))
            continue

        # per-unit check loop
        for unit, unit_data in services['units'].iteritems():
            check_str = 'agent: {}'.format(unit)
            if check_str in check_hist:
                logging.info('Already checked  {}'.format(check_str))
            else:
                check_hist.add(check_str)
                if unit_data['agent-state'] == 'started':
                    logging.info('Agent OK for unit: {}'.format(unit))
                elif unit_data['agent-state'] == 'down':
                    logging.warn('Agent down for unit: {}'.format(unit))
                    sick.append('unit:  {}'.format(unit))
                else:
                    logging.warn('Agent state unknown ({}) for unit: '
                                 '{}'.format(unit_data['agent-state'],
                                             unit))
                    sick.append('unit:  {}'.format(unit))

                # connectivity checks
            if 'public-address' in unit_data.keys():
                host = unit_data['public-address']

                # SSH port connect check
                check_str = 'ssh socket: {}'.format(host)
                if check_str in check_hist:
                    logging.info('Already checked  {}'.format(check_str))
                else:
                    check_hist.add(check_str)
                    ret = jshc_tcp_socket_check(host=host)
                    if not ret[0]:
                        sick.append(ret[1])
                        sick.append('unit:  {}'.format(unit))

                # SSH command check
                check_str = 'ssh cmd: {}'.format(host)
                if check_str in check_hist:
                    logging.info('Already checked  {}'.format(check_str))
                else:
                    check_hist.add(check_str)
                    ret = jshc_ssh_command_check(host=host)
                    if not ret[0]:
                        sick.append(ret[1])
                        sick.append('unit:  {}'.format(unit))

                # ping check
                ret = jshc_ping_check(host=host, check_hist=check_hist)
                if ret[2]:
                    check_hist.add(ret[2])
                if ret[0] is False:
                    sick.extend([ret[1], 'unit:  {}'.format(unit)])
            else:
                logging.info('No public-address for unit:  {}'.format(unit))

    # per-machine unit state check
    logging.debug('Checking agents per machine.')
    for mach, mach_data in j_stat['machines'].iteritems():
        if ignore_bootstrap_node and mach == '0':
            logging.info('Ignoring machine {}'.format(mach))
            continue

        logging.info('=== Checking machine: {} ==='.format(mach))

        # agent status check
        if 'agent-state' not in mach_data.keys():
            logging.warn('No agent-state reported for {}'.format(mach))
            sick.append('machine:  {}'.format(mach))
            continue

        check_hist.add('agent: {}'.format(mach))
        if mach_data['agent-state'] == 'started':
            logging.info('Agent OK for machine: {}'.format(mach))
        elif mach_data['agent-state'] == 'down':
            logging.warn('Agent down for machine: {}'.format(mach))
            sick.append('machine:  {}'.format(mach))
        else:
            logging.warn('Agent state unknown ({}) for '
                         'machine:  {}'.format(mach_data['agent-state'],
                                               mach))
            sick.append('machine:  {}'.format(mach))

        # bootstrap node check
        if 'state-server-member-status' in mach_data:
            if mach_data['state-server-member-status'] == 'has-vote':
                found_bootstrap_node = True
                logging.info('Found bootstrap machine: {}'.format(mach))
                if 'dns-name' in mach_data.keys():
                    host = mach_data['dns-name']

                    # juju agent port connect check
                    check_str = 'agent socket: {}'.format(host)
                    if check_str in check_hist:
                        logging.info('Already checked  {}'.format(check_str))
                    else:
                        check_hist.add(check_str)
                        ret = jshc_tcp_socket_check(host=host, port=17070)
                        if not ret[0]:
                            sick.append(ret[1])
                            sick.append('machine:  {}'.format(mach))

                    # SSH command check
                    check_str = 'ssh cmd: {}'.format(host)
                    if check_str in check_hist:
                        logging.info('Already checked  {}'.format(check_str))
                    else:
                        check_hist.add(check_str)
                        ret = jshc_ssh_command_check(host=host)
                        if not ret[0]:
                            sick.append(ret[1])
                            sick.append('machine:  {}'.format(mach))

        # connectivity checks
        if 'dns-name' in mach_data.keys():
            host = mach_data['dns-name']

            # SSH port connect check
            check_str = 'ssh socket: {}'.format(host)
            if check_str in check_hist:
                logging.info('Already checked  {}'.format(check_str))
            else:
                check_hist.add(check_str)
                ret = jshc_tcp_socket_check(host=host)
                if not ret[0]:
                    sick.append(ret[1])
                    sick.append('machine:  {}'.format(mach))

            # SSH command check
            check_str = 'ssh cmd: {}'.format(host)
            if check_str in check_hist:
                logging.info('Already checked  {}'.format(check_str))
            else:
                check_hist.add(check_str)
                ret = jshc_ssh_command_check(host=host)
                if not ret[0]:
                    sick.append(ret[1])
                    sick.append('machine:  {}'.format(mach))

            # ping check
            ret = jshc_ping_check(host=host, check_hist=check_hist)
            if ret[2]:
                check_hist.add(ret[2])
            if ret[0] is False:
                sick.extend([ret[1], 'machine:  {}'.format(unit)])

        # per-machine container state check
        if 'containers' in mach_data:
            for container in mach_data['containers']:
                check_hist.add('agent: {}'.format(container))
                if mach_data['agent-state'] == 'started':
                    logging.info('Agent OK for container: '
                                 '{}'.format(container))
                elif mach_data['agent-state'] == 'down':
                    logging.warn('Agent down for container:'
                                 ' {}'.format(container))
                    sick.append('container:  {}'.format(container))
                else:
                    logging.warn('Agent state unknown ({}) for container: '
                                 '{}'.format(mach_data['agent-state'],
                                             container))
                    sick.append('container:  {}'.format(container))

                if 'dns-name' in mach_data['containers'][container]:
                    host = mach_data['containers'][container]['dns-name']

                    # SSH port connect check
                    check_str = 'ssh socket: {}'.format(host)
                    if check_str in check_hist:
                        logging.info('Already checked  {}'.format(check_str))
                    else:
                        check_hist.add(check_str)
                        ret = jshc_tcp_socket_check(host=host)
                        if not ret[0]:
                            sick.append(ret[1])
                            sick.append('container: '.format(container))

                    # SSH command check
                    check_str = 'ssh cmd: {}'.format(host)
                    if check_str in check_hist:
                        logging.info('Already checked  {}'.format(check_str))
                    else:
                        check_hist.add(check_str)
                        ret = jshc_ssh_command_check(host=host)
                        if not ret[0]:
                            sick.append(ret[1])
                            sick.append('container:  {}'.format(container))

                    # ping check
                    ret = jshc_ping_check(host=host, check_hist=check_hist)
                    if ret[2]:
                        check_hist.add(ret[2])
                    if ret[0] is False:
                        sick.extend([ret[1], 'container:  {}'.format(unit)])

    # check results
    logging.debug('Check history: {}'.format(check_hist))
    if not found_bootstrap_node and not ignore_bootstrap_node:
        sick.append('bootstrap node (not identified)')

    logging.debug('Sick data: {}'.format(sick))
    if sick:
        sick.sort()
        logging.error('One or more units are not ok:'
                      '\n    {}'.format('\n    '.join(list(set(sick)))))
        sys.exit(1)
    else:
        logging.info('  OK: All units and machines have agents '
                     'running and they are reachable.')


def main():
    '''Define and handle command line parameters
    '''
    # Define command line options
    parser = optparse.OptionParser(USAGE)
    parser.add_option("-d", "--debug",
                      help="Enable debug output",
                      dest="debug", action="store_true", default=False)

    parser.add_option("-q", "--quiet",
                      help="Less output (WARN and ERROR only)",
                      dest="quiet", action="store_true", default=False)
    parser.add_option("--Xb",
                      help="Ignore bootstrap node (do not check it)",
                      dest="ignore_bootstrap_node", action="store_true",
                      default=False)

    params = parser.parse_args()
    (opts, args) = params

    # Handle parameters, inform user
    if opts.debug and not opts.quiet:
        logging.basicConfig(level=logging.DEBUG)
        logging.info('Logging level set to DEBUG!')
        logging.debug('parse opts: \n{}'.format(
            yaml.dump(vars(opts), default_flow_style=False)))
        logging.debug('parse args: {}'.format(args))
    elif opts.quiet:
        logging.basicConfig(level=logging.WARN)
    else:
        logging.basicConfig(level=logging.INFO)

    logging.info('Ubuntu OSCI Juju Status Health Check')
    juju_status_health_check(ignore_bootstrap_node=opts.ignore_bootstrap_node)


if __name__ == '__main__':
    main()
