#!/usr/bin/env python
'''
Find a target containing a specified string in a specified
makefile and execute it.
'''

import collections
import optparse
import logging
import os
import subprocess
import sys
import yaml

sys.path.insert(1, os.path.join(sys.path[0], '..'))
from common import osci_utils as u  # noqa

USAGE = '''Usage: %prog [options] MAKEFILENAME STRING1 [STRING2 STRING3]

Default behavior:
  Find a target containing a specified string in a specified
  makefile and execute 'make' on the first matching target name found.
  As target names vary across projects (charms), this wrapper aims to
  help with that.  String argument order is not significant.

ex:

  %prog -d Makefile @flake8

    Inspect the specified Makefile and execute make on the
    target which contains "pylint -rn" -- with debug output.
    Assumes exactly 1 match.

  %prog -d Makefile "nosetests" "unit.test"

    Same as above, but search for multiple strings.  First match wins.
'''


def makefile_to_dict(file_name):
    '''A rough little Makefile parser, to return a dictionary
    of make targets and subsequent commands.
    '''

    # Read the Makefile into a list of lines
    lines = u.read_file(file_name)
    if not lines:
        return False

    # Rough Makefile parser
    found_target, found_cmd = False, False
    _d = collections.OrderedDict()
    cmds = []
    for this_line in lines:
        line = this_line.strip()
        target = False
        if len(line) != 0 and line[0] != '#':
            if this_line[0] != '\t' and ':' in line:
                # Makefile target name found
                target = line.split(':')[0]
                cmds = []
                found_target = True
            elif this_line[0] == '\t':
                # Makefile command found
                cmds.append(line)
                found_cmd = True
        if target:
            # Add to dict with target name as key
            _d[target] = cmds

    if found_target and found_cmd:
        return _d
    else:
        if not found_target:
            logging.debug('Found no make target name(s)')
        if not found_cmd:
            logging.debug('Found no target command(s)')
        return False


def make_target(target, mf_filename):
    '''Call make target'''
    cmd = 'make -f {} {}'.format(mf_filename, target)
    logging.info('command:  {}'.format(cmd))
    cmd_split = cmd.split(' ')
    cmd_run = subprocess.Popen(cmd_split,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)

    while cmd_run.poll() is None:
        _out = cmd_run.stdout.readline()
        sys.stdout.write(_out)
        sys.stdout.flush()

    cmd_retcode = cmd_run.returncode
    if cmd_retcode != 0:
        logging.error('Make target returned non-zero.')
    return cmd_retcode


def main():
    '''Define and handle command line parameters'''
    # Define command line options
    parser = optparse.OptionParser(USAGE)
    parser.add_option("-d", "--debug",
                      help="Enable debug logging",
                      dest="debug", action="store_true", default=False)
    parser.add_option('-X', '--no-execute',
                      help='Do not execute.',
                      dest='no_exec', action='store_true', default=False)
    params = parser.parse_args()
    (opts, args) = params

    # Handle parameters, inform user
    if opts.debug:
        logging.basicConfig(level=logging.DEBUG)
        logging.info('Logging level set to DEBUG!')
        logging.debug('parse opts: \n{}'.format(
            yaml.dump(vars(opts), default_flow_style=False)))
        logging.debug('parse args: {}'.format(args))
    else:
        logging.basicConfig(level=logging.INFO)
    logging.info('Workspace dir: {}'.format(u.WORKSPACE))

    # Validate arguments
    if len(args) < 2:
        logging.error('Expected 2 or more arguments:  a Makefile name '
                      'and one or more search strings.')
        parser.print_help()
        sys.exit(1)

    # Read Makefile into an ordered dictionary
    mf_filename = args[0]
    logging.info('Reading file: {}'.format(mf_filename))
    mf_data = makefile_to_dict(mf_filename)
    if not mf_data:
        logging.error('Makefile missing or invalid: {}'.format(mf_filename))
        sys.exit(3)
    else:
        logging.debug('Found {} make targets in {}.'.format(len(mf_data),
                                                            mf_filename))
        logging.debug('Makefile dict data:\n{}'.format(
            yaml.dump(mf_data, default_flow_style=False)))

    # Search Makefile targets for matching string in commands
    mf_sstrings = args[1:]
    logging.info('Searching for: {}'.format(mf_sstrings))
    tgt_matches = []
    for target, cmds in mf_data.iteritems():
        for cmd in cmds:
            for mf_sstring in mf_sstrings:
                if mf_sstring in cmd and target not in tgt_matches:
                    tgt_matches.append(target)
                    break

    # Handle matches
    if tgt_matches:
        num_matches = len(tgt_matches)
        logging.debug('Found {} target match(es), first '
                      'found wins:  {}'.format(num_matches, tgt_matches[0]))
        logging.debug('Matches:\n{}'.format(tgt_matches))

        if opts.no_exec:
            logging.info('Not executing due to options.')
            logging.info('Would have made: {}'.format(tgt_matches[0]))
        else:
            # Make it!
            ret = make_target(tgt_matches[0], mf_filename)
            logging.debug(ret)
            if ret != 0:
                sys.exit(1)
    else:
        logging.info('Search string not found in makefile target commands.')
        logging.error('No make target was executed.')
        sys.exit(2)


if __name__ == '__main__':
    main()
