#!/usr/bin/python
'''Gathers info about nova instances, filtered by instance name based on
local default juju environment name, and writes a summary yaml text file.
'''

import logging
import os
import sys
from novaclient import client as nv_client

sys.path.insert(1, os.path.join(sys.path[0], '..'))
from common import osci_utils as u  # noqa

# Set default logging level
logging.basicConfig(level=logging.INFO)


def summarize_nova_instances(creds, name_substr=''):
    '''Gather info about nova instances, optionally filtered by
    instance name substring, and return summary data in a dictionary.
    '''
    nova = nv_client.Client('2', **creds[1])
    servers = nova.servers.list()
    nova_ddata = {}
    logging.info('Instance name filter substring: {}'.format(name_substr))
    inst_count = 0

    if not name_substr:
        name_substr = ''

    for server in servers:
        if name_substr in str(server.name):
            inst_count += 1

            try:
                image = nova.images.get(server.image['id'])
                img_name = str(image.name)
                img_date = str(image.created)
            except nv_client.exceptions.NotFound:
                # It is possible for an instance to be based on an image
                # which no longer exists.
                img_name = 'NotFound'
                img_date = 'NotFound'

            nova_ddata[str(server.id)] = {
                'inst-name': str(server.name),
                'inst-date': str(server.created),
                'inst-stat': str(server.status),
                'inst-flav': str(server.flavor['id']),
                'img-id': str(server.image['id']),
                'img-name': img_name,
                'img-date': img_date,
                'net': u.dict_unicode_to_str(server.addresses)
            }

    logging.info('Instance count (all): {}'.format(len(servers)))
    logging.info('Instance count (filtered): {}'.format(inst_count))
    return nova_ddata


def write_instance_summary_file(inst_summs):
    '''Write out instance summary data to yaml txt file
    '''
    out_file = 'nova-instance-summary.{}.txt'.format(u.BUILD_NUMBER)
    target = os.path.join(u.WORKSPACE, out_file)
    logging.info(' + Creating nova instance summary file: {}'.format(target))
    u.write_yaml(inst_summs, target)

    # Quick confirmation of file success
    test_file = u.read_yaml(target)
    if len(inst_summs) != len(test_file):
        msg = ' ! Instance summary file data may not be complete.'
        return msg


def main():
    '''Gather nova info about current nova instances, create summary file.
    '''

    logging.info('UOSCI Nova Instance Summarizer')
    logging.info('WORKSPACE dir: {}'.format(u.WORKSPACE))

    creds = u.get_os_creds()
    if not creds:
        logging.error('Did you source novarc?')
        sys.exit(1)

    # get summary of matching nova instances
    inst_summs = summarize_nova_instances(creds, name_substr=u.JENV_NAME)
    logging.debug('Instance summary data: \n{}'.format(
        u.dict_to_yaml(inst_summs)))

    # write summary to file in workspace
    ret = write_instance_summary_file(inst_summs)
    if ret:
        raise ValueError(ret)


if __name__ == '__main__':
    main()
