#!/usr/bin/env python
'''
UOSCI Launchpad Branch Scanner
'''

import logging
import os
import sys
from copy import deepcopy

sys.path.insert(1, os.path.join(sys.path[0], '..'))
from common import osci_utils as u  # noqa

USAGE = '''Usage: %prog [options]

UOSCI Launchpad Branch Scanner

Scans trigger-branches from the mappings.yaml data, watching for
changes to the launchpad branches, while keeping a simple state file.

Creates deploy, unit, lint, amulet properties files for consumption
by Jenkins parameterized builds.

Assumes ubuntu-openstack-packagesource and other conventions used by
bundles in lp:openstack-charm-testing.

Amulet tests and bundle deploy tests will be automatically suppressed
for any non-charm branches in the trigger list.

Optionally handles per-branch overrides via hash tags appended to
branch strings in the mappings.yaml control file.

    ie.   - lp:ubuntu-openstack-ci#NOUNIT
    ie.   - lp:charms/trusty/ubuntu#NODEPLOY#NOUNIT

    #NODEPLOY   Branch changes will not trigger deploy tests.
    #NOAMULET   Branch changes will not trigger amulet tests.
    #NOUNIT     Branch changes will not trigger unit tests.
    #NOLINT     Branch changes will not trigger lint tests.

Example usage:

    %prog
  Assumes True for amulet, lint, deploy and all other
  trigger types and branch scanning routines, and creates
  properties files accordingly.

    %prog -Nd
  Performs all default actions (with debug output), leaving the
  state file as it was found.

    %prog -FN --XD --XA
  Forces creation of trigger properties files, only for lint,
  without scanning any of the branches, and without altering
  the state file.

    %prog --XD
  Scans branches and creates trigger properties files for
  all checks EXCEPT deployments.
'''


def get_changed_branches(conf_map, opts):
    '''Query previous and current branch data, compare it, optionally
    override it, write it when appropriate, and return it.'''

    # Determine previous state
    state_fn = u.get_state_file_name(conf_map)
    state_prev = u.read_state_file(state_fn)

    # Determine current state
    if opts.no_scan:
        logging.info('Not scanning branches, forcing to a not-changed state.')
        state_now = deepcopy(state_prev)
    elif opts.force:
        logging.info('Forcing all branches to a simulated changed state.')
        state_now = u.fake_branch_states(conf_map['trigger-branches'])
    else:
        logging.info('Querying interwebs for branch states.')
        state_now = u.query_branch_states(conf_map['trigger-branches'])

    # Compare current branch state with previous state
    comps = u.dict_compare(state_prev, state_now)
    changed = list(set(comps['added'] + comps['diff']))
    removed = comps['rmvd']
    changed_fn = '{}.changed'.format(state_fn)

    # Handle changes, if any
    if changed or removed:
        logging.info('Changed: \n{}'.format(u.dict_to_yaml(changed)))
        logging.info('Removed: \n{}'.format(u.dict_to_yaml(removed)))
        if not opts.dry and not opts.no_state:
            u.write_state_file(state_now, state_fn)
            u.write_changed_file(changed, changed_fn)
        return changed
    else:
        logging.info('No branch changes detected.  Good day.')
        if not opts.dry and not opts.no_state:
            if os.path.isfile(changed_fn):
                os.remove(changed_fn)
            return None


def main():
    '''Go!'''
    # Initialize configuration and command line options
    (opts, args) = u.define_scanner_command_options(USAGE)
    u.handle_scanner_command_options(opts, args)
    conf_map = u.read_yaml(u.MAP_FILE)

    # Scan branches for changes and hash tags, optionally write data files
    changed = get_changed_branches(conf_map, opts)
    tagged = u.get_tagged_branches(conf_map)

    # Determine if deploys should be triggered
    deploy_is_needed = u.is_deploy_needed(changed, tagged,
                                          blacktags=['NODEPLOY'])

    # Construct target dict data for template consumption
    if changed and opts.deploy and deploy_is_needed:
        deploy_targets = u.create_deploy_properties_dict(
            conf_map['trigger-deployments'])
        logging.debug('Deploy targets: \n{}'.format(
            u.dict_to_yaml(deploy_targets)))

    if changed and opts.amulet:
        amulet_targets = u.create_properties_dict(changed, tagged,
                                                  blacktags=['NOAMULET'],
                                                  charms_only=True)
        logging.debug('Amulet targets: \n{}'.format(
            u.dict_to_yaml(amulet_targets)))

    if changed and opts.lint:
        lint_targets = u.create_properties_dict(changed, tagged,
                                                blacktags=['NOLINT'])
        logging.debug('Lint targets: \n{}'.format(
            u.dict_to_yaml(lint_targets)))

    if changed and opts.unit:
        unit_targets = u.create_properties_dict(changed, tagged,
                                                blacktags=['NOUNIT'])
        logging.debug('Unit targets: \n{}'.format(
            u.dict_to_yaml(unit_targets)))

    # Create properties files for use by Jenkins parameterized builds
# common
    if not opts.dry:
        if changed and opts.deploy and deploy_is_needed:
            logging.info('Creating deploy-properties files')
            u.create_properties_files(deploy_targets,
                                      source='deploy-properties.tmpl',
                                      out_file='deploy-properties.{}')

        if changed and opts.amulet:
            logging.info('Creating amulet-properties files')
            u.create_properties_files(amulet_targets,
                                      source='amulet-properties.tmpl',
                                      out_file='amulet-properties.{}')

        if changed and opts.lint:
            logging.info('Creating lint-properties files')
            u.create_properties_files(lint_targets,
                                      source='lint-properties.tmpl',
                                      out_file='lint-properties.{}')

        if changed and opts.unit:
            logging.info('Creating unit-properties files')
            u.create_properties_files(unit_targets,
                                      source='unit-properties.tmpl',
                                      out_file='unit-properties.{}')


if __name__ == '__main__':
    main()
