#!/usr/bin/env python
'''
UOSCI Launchpad Merge Proposal Scanner
'''

import logging
import os
import sys
import time
import traceback
from copy import deepcopy

sys.path.insert(1, os.path.join(sys.path[0], '..'))
from common import osci_utils as u  # noqa

USAGE = '''Usage: %prog [options]

UOSCI Launchpad Merge Proposal Scanner

Scans trigger-branches from the mappings.yaml data, watching for
status of merge proposals on launchpad branches, while keeping a
simple state file.

Creates deploy, unit, lint, amulet properties files for consumption
by Jenkins parameterized builds.

Assumes ubuntu-openstack-packagesource and other conventions used by
bundles in lp:openstack-charm-testing.

Amulet tests and bundle deploy tests will be automatically suppressed
for any non-charm branches in the trigger list.

Optionally handles per-branch overrides via hash tags appended to
branch strings in the mappings.yaml control file.

    ie.   - lp:ubuntu-openstack-ci#NOUNIT
    ie.   - lp:charms/trusty/ubuntu#NODEPLOY#NOUNIT

    #NODEPLOY   Branch changes will not trigger deploy tests.
    #NOAMULET   Branch changes will not trigger amulet tests.
    #NOUNIT     Branch changes will not trigger unit tests.
    #NOLINT     Branch changes will not trigger lint tests.

Example usage:

    %prog
  Assumes True for amulet, lint, deploy and all other
  trigger types and branch scanning routines, and creates
  properties files accordingly.

    %prog -Nd
  Performs all default actions (with debug output), leaving the
  state file as it was found.

    %prog -FN --XD --XA
  Forces creation of trigger properties files, only for lint,
  without scanning any of the branches, and without altering
  the state file.

    %prog --XD
  Scans branches and creates trigger properties files for
  all checks EXCEPT deployments.
'''


def _get_branch_info(branch):
    '''Returns a dict containing the branch information.

    The branch information includes keys for:
        display - branch display name
        web - branch web link
        owner - branch owner

    Any exception handling which may occur is intentionally
    deferred to the caller.

    :param branch: the launchpadlib branch reference
    :return: a dict containing the branch information
    '''
    return {
        'display': str(branch.display_name),
        'web': str(branch.web_link),
        'owner': str(branch.owner.name)
    }


def _comment_author_meets_requirements(comment, karma_min=100):
    '''Basic safety and sanity checks against comment author.'''
    meets_requirements = True
    if not comment.author.is_ubuntu_coc_signer:
        logging.debug('Comment author has not signed the '
                      'Ubuntu Code of Conduct')
        meets_requirements = False

    if comment.author.is_probationary:
        logging.debug('Comment author is LP probationary')
        meets_requirements = False

    if comment.author.karma < karma_min:
        logging.debug('Comment author has low LP karma')
        meets_requirements = False

    return meets_requirements


def get_last_recheck_timestamp(mp, launchpad):
    '''Returns the timestamp of the last recheck request.

    Rechecks are specific comments (#recheck) left against a merge
    proposal to re-trigger the tests.

    :mp: the launchpadlib merge proposal to scan
    :launchpad: the launchpadlib connection object
    :returns: datetime.datetime of the latest matching request
    '''

    # Half-prepped to turn this into a multi-purpose handler if need be
    mp_command = "#recheck"

    # Get matching comments posted after the src branch was last modified
    comment_matches = []
    last_modified_date = mp.source_branch.date_last_modified
    for comment in mp.all_comments:

        comment_timestamp = comment.date_created
        if comment_timestamp < last_modified_date:
            continue

        comment_msg_body = str(comment.message_body).strip()
        if comment_msg_body.lower().startswith(mp_command.lower()):
            comment_matches.append(comment)

    if not comment_matches:
        return None

    # Get timestamp of latest matching mp command
    comment = comment_matches[-1]
    comment_author = '{} ({})'.format(comment.author.display_name,
                                      comment.author.name)

    logging.info('{} requested {} @ '
                 '{}'.format(comment_author,
                             mp_command,
                             comment_timestamp))

    if _comment_author_meets_requirements(comment):
        return comment.date_created
    else:
        logging.warn('Request denied: author does not meet requirements')
        return None


def query_branch_mp_states(launchpad, branch_list,
                           mp_status_filter, max_tries=10):
    '''Get current state of merge proposals on a list of branches,
    filtered on a list of proposal statuses.  Return a dict.'''
    state = {}
    for branch_str in branch_list:
        logging.info('Checking MPs on {}'.format(branch_str))
        proposals = u.lpl_get_mps(launchpad,
                                  lp_branch_url=branch_str,
                                  status=mp_status_filter)
        if not proposals:
            continue

        for mp in proposals:
            this = {}
            mp_str = str(mp.address)
            mp_web = str(mp.web_link)
            mp_status = str(mp.queue_status)
            mp_registrant = str(mp.registrant.name)

            this['scanned-branch'] = branch_str
            this['mp-web'] = mp_web
            this['mp-status'] = mp_status
            this['mp-registrant'] = mp_registrant

            logging.debug('Merge proposal state: {} ({}) \n'
                          '  {} {}'.format(mp_status, mp_registrant,
                                           mp_str, mp_web))
            try:
                this['src'] = _get_branch_info(mp.source_branch)
                this['tgt'] = _get_branch_info(mp.target_branch)
            except Exception:
                logging.error(traceback.format_exc())
                src_or_tgt = 'source' if 'src' in this else 'target'
                logging.error('Skipping {}\nCould not retrieve {} branch '
                              'info'.format(mp.web_link, src_or_tgt))
                u.touch(os.path.join(u.WORKSPACE, 'fyi.branch.skipped'))
                logging.error('Could be a private LP branch, check perms.')
                continue

            last_modified_date = mp.source_branch.date_last_modified

            this['dates'] = {
                'src-branch-last-mod': str(last_modified_date),
                'created': str(mp.date_created),
                'merged': str(mp.date_merged),
                'requested': str(mp.date_review_requested),
                'reviewed': str(mp.date_reviewed)
            }

            # Check for recheck commands within mp comments
            last_recheck_timestamp = get_last_recheck_timestamp(mp, launchpad)
            if last_recheck_timestamp:
                this['dates'].update({
                    'recheck': str(last_recheck_timestamp)
                })

            state[mp_str] = this
    return state


def get_changed_mp_state(conf_map, opts):
    '''Scan branches for merge proposals, compare with last state if
    known, and return a list of changed MPs and current MP state.'''

    # initialize launchpad connection
    if opts.force or opts.no_scan:
        logging.debug('Not connecting to Launchpad API due to options')
        launchpad = None
    else:
        logging.debug('Connecting Launchpad API...')
        launchpad = u.lpl_connect(anonymous=False)

    # Determine previous state
    mp_state_fn = u.get_state_file_name(conf_map, key='mp-state-file')
    mp_state_prev = u.read_state_file(mp_state_fn)

    # Determine current state
    if opts.no_scan:
        logging.info('Not scanning branches, forcing to a not-changed state.')
        state = deepcopy(mp_state_prev)
    elif opts.force:
        logging.info('Forcing all branches to a simulated changed state.')
        state = deepcopy(mp_state_prev)
        fake = 'SIMULATED CHANGE {}'.format(time.time())
        for k in state.keys():
            state[k]['dates']['src-branch-last-mod'] = fake
    else:
        logging.info('Querying LP for branch merge proposals and states.')
        branch_list = u.list_strip_tags(conf_map['trigger-branches'])
        mp_status_filter = conf_map['mp-status-filter']
        state = query_branch_mp_states(launchpad, branch_list,
                                       mp_status_filter, max_tries=10)

    # Compare current MP state with previous state
    comps = u.dict_compare(mp_state_prev, state)
    changed = list(set(comps['added'] + comps['diff']) - set(comps['rmvd']))
    removed = comps['rmvd']
    mp_changed_fn = '{}.changed'.format(mp_state_fn)

    # Handle changes, if any
    #   Refactor to common utils?:
    if changed or removed:
        logging.info('Changed: \n{}'.format(u.dict_to_yaml(changed)))
        logging.info('Removed: \n{}'.format(u.dict_to_yaml(removed)))
        if not opts.dry and not opts.no_state:
            u.write_state_file(state, mp_state_fn)
            u.write_changed_file(changed, mp_changed_fn)
    else:
        logging.info('No MP changes detected.  Good day.')
        if not opts.dry and not opts.no_state:
            if os.path.isfile(mp_changed_fn):
                os.remove(mp_changed_fn)

    return changed, state


def create_deploy_properties_dict(changed, state, conf_map,
                                  tagged={}, blacktags=[]):
    '''Generate properties data Ubuntu + OpenStack + Install Source
    + Charm Set, return dictionary which can be passed to a
    make deploy-properties files routine.
    '''
    rls_supp = u.get_u_supported()
    contexts = []

    for mp_src in changed:
        branch_str = state[mp_src]['scanned-branch']
        mp_branch_str = state[mp_src]['src']['display']
        mp_reg = state[mp_src]['mp-registrant']
        mp_num = u.strip_chars(mp_src.split('@')[0])
        is_charm = u.is_charm_branch(branch_str)
        display_name = '{} {}'.format(mp_reg, mp_num)

        if not is_charm:
            continue

        for charm_set, targets in \
                conf_map['mp-trigger-deployments'].iteritems():
            targets.sort(reverse=True)
            for target in targets:
                _split = target.split('-')
                u_rls = _split[0]
                os_rls = _split[1]

                if len(_split) == 3:
                    inst_src = _split[2]
                else:
                    inst_src = 'default'

                if u_rls not in rls_supp:
                    logging.warning('Detected an unsupported '
                                    'release: {}'.format(u_rls))

                contexts.append({
                    'charm_set': charm_set,
                    'u_rls': u_rls,
                    'os_rls': os_rls,
                    'inst_src': inst_src,
                    'display_name': display_name,
                    'mp_web': state[mp_src]['mp-web'],
                    'mp_src_br': mp_branch_str,
                    'mp_tgt_br': branch_str
                })
    return contexts


def create_properties_dict(changed, state, tagged={},
                           blacktags=[], charms_only=False):
    '''Generate context dict for creation of lint-properties,
    unit-properties, and amulet-properties files.
    '''
    contexts = []
    changed.sort()
    for mp_src in changed:
        branch_str = state[mp_src]['scanned-branch']
        mp_branch_str = state[mp_src]['src']['display']
        mp_reg = state[mp_src]['mp-registrant']
        mp_num = u.strip_chars(mp_src.split('@')[0])
        is_charm = u.is_charm_branch(branch_str)
        is_next = u.is_next_branch(branch_str)

        # Skip branches tagged with blacktags
        if u.is_tagged(branch_str, tagged, blacktags):
            continue

        if is_charm and is_next:
            base_name = u.get_next_base_name(branch_str)
            display_name = '{}-next for {} {}'.format(base_name,
                                                      mp_reg, mp_num)
        else:
            base_name = u.get_stable_base_name(branch_str)
            display_name = '{} for {} {}'.format(base_name,
                                                 mp_reg, mp_num)

        if (is_charm and charms_only) or not charms_only:
            contexts.append({
                'branch': mp_branch_str,
                'base_name': base_name,
                'display_name': display_name,
                'mp_web': state[mp_src]['mp-web'],
                'mp_src_br': mp_branch_str,
                'mp_tgt_br': branch_str
            })
    return contexts


def main():
    '''Define and handle command line parameters
    '''
    # Initialize configuration and command line options
    (opts, args) = u.define_scanner_command_options(USAGE)
    u.handle_scanner_command_options(opts, args)
    conf_map = u.read_yaml(u.MAP_FILE)

    # Scan branches for changes and hash tags, optionally write data files
    tagged = u.get_tagged_branches(conf_map)
    changed, state = get_changed_mp_state(conf_map, opts)

    # Construct target dict data for template consumption
    if changed and opts.deploy:
        deploy_targets = create_deploy_properties_dict(changed, state,
                                                       conf_map, tagged,
                                                       blacktags=['NODEPLOY'])
        logging.debug('Deploy targets: \n{}'.format(
            u.dict_to_yaml(deploy_targets)))

    if changed and opts.amulet:
        amulet_targets = create_properties_dict(changed, state,
                                                tagged, blacktags=['NOAMULET'],
                                                charms_only=True)
        logging.debug('Amulet targets: \n{}'.format(
            u.dict_to_yaml(amulet_targets)))

    if changed and opts.lint:
        lint_targets = create_properties_dict(changed, state,
                                              tagged, blacktags=['NOLINT'])
        logging.debug('Lint targets: \n{}'.format(
            u.dict_to_yaml(lint_targets)))

    if changed and opts.unit:
        unit_targets = create_properties_dict(changed, state,
                                              tagged, blacktags=['NOUNIT'])
        logging.debug('Unit targets: \n{}'.format(
            u.dict_to_yaml(unit_targets)))

    # Create properties files for use by Jenkins parameterized builds
# common
    if not opts.dry:
        if changed and opts.deploy:
            logging.info('Creating deploy-properties files')
            u.create_properties_files(deploy_targets,
                                      source='deploy-properties.tmpl',
                                      out_file='deploy-properties.{}')

        if changed and opts.amulet:
            logging.info('Creating amulet-properties files')
            u.create_properties_files(amulet_targets,
                                      source='amulet-properties.tmpl',
                                      out_file='amulet-properties.{}')

        if changed and opts.lint:
            logging.info('Creating lint-properties files')
            u.create_properties_files(lint_targets,
                                      source='lint-properties.tmpl',
                                      out_file='lint-properties.{}')

        if changed and opts.unit:
            logging.info('Creating unit-properties files')
            u.create_properties_files(unit_targets,
                                      source='unit-properties.tmpl',
                                      out_file='unit-properties.{}')


if __name__ == '__main__':
    main()
