#!/usr/bin/env python
'''
UOSCI Mojo spec group trigger tool
'''

import distro_info
import logging
import optparse
import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], '..'))
from common import osci_utils as u  # noqa

USAGE = '''Usage: %prog [options]

UOSCI Mojo spec group trigger tool

Example usage:

%prog
    Create properties files for all combinations of Ubuntu + Openstack
    against all specs in all groups listed in the mojo-spec-groups
    section of the mappings.yaml file.

%prog -d0
    Same as above, but with debug output, and do not actually create
    the properties files.

%prog -H
    Create properties files for "hot" combinations of Ubuntu + Openstack
    against all specs in all groups listed in the mojo-spec-groups
    section of the mappings.yaml file.

%prog -d -g full_stack
    Create properties files for all combinations of Ubuntu + Openstack
    against all specs in the full_stack group.

%prog -d -g object_storage -o kilo
    Create properties files for all combinations of Ubuntu + Kilo
    against all specs in the object_storage group.

%prog -d -Ls next
    Create properties files for all U+OS combos, against all specs
    which contain "next" in the spec name.

%prog -d -s next_deploy_baremetal
    Create properties files for all U+OS combos, against all specs
    named "next_deploy_baremetal" (exact spec name match).

%prog -d -g full_stack -Ls charmhelper
    Create properties files for all U+OS combos, against all specs
    in the full_stack group with "charmhelper" in the spec name.

%prog -d -u trusty -o kilo
    Create properties files for Trusty + Kilo, against all specs
    in all groups.

'''


def define_scanner_command_options(usage_string):
    '''Define command line options.'''
    parser = optparse.OptionParser(usage_string)
    parser.add_option('-d', '--debug',
                      help='Enable debug logging',
                      dest='debug', action='store_true', default=False)
    parser.add_option('-0', '--dry-run',
                      help='Do not create properties files, and do not '
                           'update state file(s).',
                      dest='dry', action='store_true', default=False)
    parser.add_option('-H', '--hot-only',
                      help='Only process specs for "hot" combos.',
                      dest='hot_only', action='store_true', default=False)
    parser.add_option('-g', '--group-name',
                      help='Only process the specific group name provided.',
                      dest='group_name', type='string', action='store')
    parser.add_option('-s', '--spec-name',
                      help='Only process spec(s) which exactly match '
                           'the name provided.',
                      dest='spec_name', type='string', action='store')
    parser.add_option('-L', '--loose-match',
                      help='Optionally use with -s to match substrings of '
                           'spec names as well as full spec names.',
                      dest='loose_match', action='store_true', default=False)
    parser.add_option('-u', '--ubuntu-release',
                      help='Only process spec(s) for the Ubuntu release '
                           'named (lower case).',
                      dest='u_release', type='string', action='store')
    parser.add_option('-o', '--openstack-release',
                      help='Only process spec(s) for the OpenStack '
                           'release named (lower case).',
                      dest='os_release', type='string', action='store')
    return parser.parse_args()


def handle_scanner_command_options(opts, args):
    '''Handle command options and inform user.'''
    if opts.debug:
        logging.basicConfig(level=logging.DEBUG)
        logging.info('Logging level set to DEBUG!')
        logging.debug('parse opts: \n{}'.format(u.dict_to_yaml(vars(opts))))
        logging.debug('parse args: {}'.format(args))
    else:
        logging.basicConfig(level=logging.INFO)

    logging.info('Workspace dir: {}'.format(u.WORKSPACE))

    if opts.dry:
        logging.info(' - Dry run enabled, properties files '
                     'will not be created.')

    if opts.hot_only:
        logging.info(' - Only processing "hot" combos.')

    if opts.hot_only and (opts.u_release or opts.os_release):
        logging.error('Hot -H may not be used with -u or -o options.')
        sys.exit(1)

    if opts.group_name:
        logging.info(' - Only processing group:  {}'.format(opts.group_name))

    if opts.spec_name:
        logging.info(' - Filtering on spec name: '
                     '{}'.format(opts.spec_name))

    if opts.loose_match:
        logging.info(' - Loose match enabled, spec name filter will match '
                     'substrings as well as full spec names.')

    if opts.loose_match and not opts.spec_name:
        logging.error('Use of -L loose match also requires -s.')
        sys.exit(1)


# Helpers
def get_ubuntu_releases(include_all=False):
    """Return a dict of supported, lts, and Ubuntu releases
    in order of release."""
    _d = distro_info.UbuntuDistroInfo()
    releases = {}
    releases['lts'] = []
    releases['supported'] = _d.supported()
    if include_all:
        releases['all'] = _d.all

    # Build lts list, as _d.lts() only returns the most recent lts
    for release in releases['supported']:
        if _d.is_lts(release):
            releases['lts'].append(release)

    logging.debug('Ubuntu releases:\n{}'.format(u.dict_to_yaml(releases)))
    return releases


def validate_u_os_map(u_os_releases, u_releases):
    '''Validate mapped Ubuntu releases as supported.  Also check that
    all supported Ubuntu releases are mapped.'''
    msg = []

    # Ensure all mapped Ubuntu releases are supported
    u_bad = []
    for release in u_os_releases.keys():
        if release not in u_releases['supported']:
            u_bad.append(release)

    if u_bad:
        msg.append('Mapped Ubuntu releases are not supported: '
                   '{}'.format(u_bad))
    else:
        logging.debug('All mapped Ubuntu releases are supported.')

    # Ensure all supported Ubuntu releases are mapped
    not_mapped = []
    for release in u_releases['supported']:
        if release not in u_os_releases.keys():
            not_mapped.append(release)

    if not_mapped:
        msg.append('Not all supported Ubuntu releases are mapped. '
                   'Missing: {}'.format(not_mapped))
    else:
        logging.debug('All supported Ubuntu releases are mapped.')

    if msg:
        logging.debug('\n'.join(msg))
        return ' '.join(msg)
    else:
        return None


def create_mojo_properties_dict(mojo_spec_groups, os_u_releases,
                                u_release_filter=None, os_release_filter=None,
                                hot_filter=None):
    '''Generate properties data which can be passed to a make mojo runner
    properties files for consumption by Jenkins parameterized jobs, computed
    based on data in the mappings yaml.
    '''
    contexts = []
    logging.debug('Hot filter:\n{}'.format(u.dict_to_yaml(hot_filter)))
    for group, spec_names in mojo_spec_groups.items():
        for spec_name in spec_names:
            for os_release, u_releases in os_u_releases.items():
                if os_release_filter and os_release != os_release_filter:
                    logging.debug('Ignoring {} for {} due to '
                                  'OpenStack release name'
                                  'filter.'.format(os_release, spec_name))
                    continue
                if hot_filter and os_release not in hot_filter.keys():
                    logging.debug('Ignoring {} for {} due to hot '
                                  'filter.'.format(os_release, spec_name))
                    continue
                for u_release in u_releases:
                    if u_release_filter and u_release != u_release_filter:
                        logging.debug('Ignoring {} for {} due to '
                                      'Ubuntu release name '
                                      'filter.'.format(u_release, spec_name))
                        continue
                    if hot_filter and u_release not in hot_filter[os_release]:
                        logging.debug('Ignoring {} for {} due to hot '
                                      'filter.'.format(u_release, spec_name))
                        continue
                    mojo_series = u_release
                    mojo_spec = 'specs/{}/{}/{}'.format(group,
                                                        spec_name,
                                                        os_release)
                    display_name = '{} @ {}'.format(mojo_spec,
                                                    mojo_series)
                    contexts.append({
                        'mojo_spec': mojo_spec,
                        'mojo_series': mojo_series,
                        'display_name': display_name
                    })
    return contexts


def filter_spec_names(mojo_spec_groups, spec_name_filter, loose=False):
    '''Return dict of groups and spec names, where the spec name matches
    a provided string filter.'''
    _d = {}
    for group, spec_names in mojo_spec_groups.items():
        for spec_name in spec_names:
            if spec_name == spec_name_filter or \
                    (loose and spec_name_filter in spec_name):
                logging.info('Found spec based on name filter: '
                             '{}/{}'.format(group, spec_name))
                if group not in _d.keys():
                    _d[group] = []
                _d[group].append(spec_name)
            else:
                logging.debug('Ignoring spec: {}/{}'.format(group, spec_name))
    return _d


def main():
    '''Go!'''
    # Initialize configuration and command line options
    (opts, args) = define_scanner_command_options(USAGE)
    handle_scanner_command_options(opts, args)
    conf_map = u.read_yaml(u.MAP_FILE)

    # Handle group options
    if opts.group_name and opts.group_name in conf_map['mojo-spec-groups']:
        mojo_spec_groups = {
            opts.group_name: conf_map['mojo-spec-groups'][opts.group_name]
        }
    elif not opts.group_name:
        mojo_spec_groups = conf_map['mojo-spec-groups']
    else:
        logging.error('Invalid group filter.  Group not found '
                      '({}).'.format(opts.group_name))
        sys.exit(1)

    # Handle name options
    if opts.spec_name:
        mojo_spec_groups = filter_spec_names(mojo_spec_groups,
                                             opts.spec_name,
                                             opts.loose_match)

    logging.debug('Mojo spec groups (filtered):\n{}'.format(
        u.dict_to_yaml(mojo_spec_groups)))

    # Get Ubuntu/OpenStack release maps
    u_releases = get_ubuntu_releases()
    u_os_releases = conf_map['u-os-map']
    os_u_releases = u.invert_dict_of_lists(u_os_releases)
    hot_combos = u.invert_dict_of_lists(conf_map['hot'])

    logging.debug('Ubuntu-to-OpenStack release map:\n{}'.format(
        u.dict_to_yaml(u_os_releases)))
    logging.debug('OpenStack-to-Ubuntu release map:\n{}'.format(
        u.dict_to_yaml(os_u_releases)))

    # Handle Ubuntu/OpenStack filters
    if opts.u_release:
        u_release_filter = opts.u_release
    else:
        u_release_filter = None

    if opts.os_release:
        os_release_filter = opts.os_release
    else:
        os_release_filter = None

    if opts.hot_only:
        hot_filter = hot_combos
    else:
        hot_filter = None

    # Validate Ubuntu-to-Openstack release map
    ret = validate_u_os_map(u_os_releases, u_releases)
    if ret:
        logging.warn(ret)

    # Construct target dict data for template consumption
    mojo_targets = create_mojo_properties_dict(mojo_spec_groups,
                                               os_u_releases,
                                               u_release_filter,
                                               os_release_filter,
                                               hot_filter)

    # Create properties files for use by Jenkins parameterized builds
    if not mojo_targets:
        logging.warn('No resultant mojo targets!  Too strict on a filter, '
                     'or no mappings.yaml data?')
    else:
        logging.debug('Mojo targets: \n{}'.format(
            u.dict_to_yaml(mojo_targets)))

        # Give feedback as to what was included
        display_names = []
        for target in mojo_targets:
            display_names.append(target['display_name'])
        logging.info('Targets:\n{}'.format(u.dict_to_yaml(display_names)))

        if not opts.dry:
            logging.info('Creating mojo-properties files')
            u.create_properties_files(mojo_targets,
                                      source='mojo-properties.tmpl',
                                      out_file='mojo-properties.{}')


if __name__ == '__main__':
    main()
