#!/usr/bin/env python
'''
Read a YAML file into a dict, write it out to the same or to a new file.
Useful in standardizing YAML format of files generated by hand for example.
'''

import optparse
import logging
import yaml
import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], '..'))
from common import osci_utils  # noqa

USAGE = '''Usage: %prog [options]

Default behavior:
  Read a YAML file into a dict, write it out to the same file
  or to a new file.  Useful in standardizing YAML syntax in
  files generated by hand.

For example:

    %prog -yd -i in.yaml -o out.yaml
'''


def main():
    '''Define and handle command line parameters
    '''
    # Define command line options
    parser = optparse.OptionParser(USAGE)
    parser.add_option("-d", "--debug",
                      help="Enable debug logging",
                      dest="debug", action="store_true", default=False)
    parser.add_option('-y', '--yes-overwrite',
                      help='Overwrite the output file',
                      dest='overwrite', action='store_true', default=False)
    parser.add_option("-i", "--in-file",
                      help="YAML input (source) file",
                      action="store", type="string", dest="in_file")
    parser.add_option("-o", "--out-file",
                      help="YAML output (destination) file",
                      action="store", type="string", dest="out_file")

    params = parser.parse_args()
    (opts, args) = params

    # Handle parameters, inform user
    if opts.debug:
        logging.basicConfig(level=logging.DEBUG)
        logging.info('Logging level set to DEBUG!')
        logging.debug('parse opts: \n{}'.format(
            yaml.dump(vars(opts), default_flow_style=False)))
        logging.debug('arg count: {}'.format(len(args)))
        logging.debug('parse args: {}'.format(args))
    else:
        logging.basicConfig(level=logging.INFO)

    # Validate options
    if not opts.in_file or not opts.out_file:
        parser.print_help()
        sys.exit(1)

    # Validate and process files
    if os.path.isfile(opts.out_file) and opts.overwrite:
        logging.warning('Output file exists and will be '
                        'overwritten: {}'.format(opts.out_file))
    elif os.path.isfile(opts.out_file) and not opts.overwrite:
        logging.warning('Output file exists and will NOT '
                        'be overwritten: {}'.format(opts.out_file))
        raise ValueError('Output file exists, overwrite option not set.')

    if not os.path.isfile(opts.in_file):
        raise ValueError('Input file not found.')

    # Read the infile
    logging.info('reading input file: {}'.format(opts.in_file))
    ddata = osci_utils.read_yaml(opts.in_file)

    # Write the outfile
    logging.info('writing out file: {}'.format(opts.out_file))
    osci_utils.write_yaml(ddata, opts.out_file)


if __name__ == '__main__':
    main()
