#!/bin/bash -e
#
# Juju 1 only - do not use with Juju >= 2
#
# WARNING: this will forcefully remove juju environment files,
# destroy all defined juju environments, and delete all nova
# instances which are believed to have been part of one of the
# defined juju environments.
echo $0

juju version

# If deleting nova instances, they must contain:
nova_instance_safety_substring="juju-osci-sv|juju-osci-lb|juju-osci-tk"

param=$1
if [ -z "$param" ]; then
  echo "Forcefully removes juju env files and destroys ALL defined"
  echo "environments.  Also wipes the current jenkins WORKSPACE!"
  echo "WARNING: Potentially dangerous, so you must confirm with YES."
  echo "Usage:  ./thisscript.sh YES"
fi

if [ "$param" != "YES" ]; then
  exit 0
fi

[[ -z "$NODE_NAME" ]] && export NODE_NAME=$(hostname)
[[ -z "$WORKSPACE" ]] && export WORKSPACE="$(mktemp -d)"
[[ -z "$OSCI_ROOT" ]] && export OSCI_ROOT="${HOME}/tools/ubuntu-openstack-ci"
[[ -z "$OPENSTACK_CHARM_TESTING_ROOT" ]] && export OPENSTACK_CHARM_TESTING_ROOT="${HOME}/tools/openstack-charm-testing"

. ${OSCI_ROOT}/run/job-parts/osci_job_common.sh
f_novarc_undercloud

# Extra safety measure, as nova delete might be called.
if [[ $OS_USERNAME != osci ]] && [[ "${JENV_NAME^^}" != *MAAS* ]]; then
  echo " ! novarc env vars do not look like OSCI, halting here."
  exit 1
fi

env_names="$(juju switch -l)"
j_env_dir="${HOME}/.juju/environments"
f_get_env_files

if [[ ! -d "$j_env_dir" ]]; then
  echo " ! Juju environment dir not found: ${j_env_dir}"
fi

# Always attempt a destroy on all enviros/models if ANY are found
if [[ -n "$env_files" ]]; then
  echo " . Forcing destroy of all Juju models"
  set -x
  for env_name in $env_names; do
    juju destroy-environment ${env_name} --force -y ||:
  done
  set +x
  echo " . End of Juju destroy"
fi

# Wait and check since Juju exits before undercloud completes destruction
f_get_env_files
f_get_my_nova_instances
tries=0
echo "Waiting up to 5 min for destroy..."
while [[ -n "$env_files" ]] || [[ -n "$my_instances" ]] && [ $tries -le 10 ]
do
  echo "Waiting for destroy..."
  sleep 30
  f_get_env_files
  f_get_my_nova_instances
  tries=$(( $tries + 1 ))
done

# Delete nova instances if matches are found
f_get_env_files
f_get_my_nova_instances

if [[ -n "$my_instances" ]]; then
  echo " ! Nova instances unexpectedly remain after juju destroy.  This is likely an undercloud issue."

  echo -e " - These nova instances will be deleted:"
  # Use filter on my_instances for added safety
  my_instances_filtered=$(echo "$my_instances" | egrep "$nova_instance_safety_substring")
  echo "$my_instances_filtered"

  while read -r line; do
    instance_id=$(echo $line | awk '{ print $2 }')
    echo " - Deleting nova instance: ${instance_id}"
    timeout 120 openstack server delete --wait $instance_id ||:
  done <<< "$my_instances_filtered"

  # Re-check nova instances
  f_get_my_nova_instances
  if [[ -n "$my_instances" ]]; then
    echo "$my_instances"
    echo " ! Nova instance cleanup failed.  One or more still exist.  Manual intervention required."
    echo "_:destroy.failed"
    # Never return if destroy failed; block to require human intervention.
    sleep infinity
  else
    echo " . Nova instances appear to be adequately cleaned up."
  fi
fi

# Delete juju .env files if they remain, after backing them up
if [[ -n "$env_files" ]]; then
  echo " ! Juju1 env file(s) remain after juju destroy:"
  echo "$env_files"
  d_str=$(date +"%Y%m%d%H%M%S%N")
  loc_dir="${HOME}/tmp.juju-env-files"
  mkdir -p $loc_dir
  tarball="env_files-${d_str}.tar.bz2"
  echo " ! Found env files in ${j_env_dir}"
  ls -al ${j_env_dir}
  echo " ! Backing them up to ${loc_dir}/${tarball}"
  tar -cjvf ${loc_dir}/${tarball} -C ${j_env_dir} .
  echo " ! Removing env files"
  rm -fv ${j_env_dir}/*env
  echo " ! Cleaning up old jenv backups"
  find ${loc_dir}/env_files*.tar.bz2 -mtime +7 -type f -delete
fi

# Confirm
JENV_NAME_SAVE="${JENV_NAME}"
echo -e "Env names:\n${env_names}"

for env_name in $env_names; do
  export JENV_NAME="${env_name}"
  juju_ok="$(timeout 30 ${OSCI_ROOT}/run/job-parts/juju1_deploy_ok.sh ||:)"

  if [[ "$juju_ok" != *CLEAR:* ]]; then
    echo " ! ${JENV_NAME} on ${NODE_NAME} is not clear."
    echo "$juju_ok"
    touch $WORKSPACE/fyi.juju-destroy.failed
    echo "_:destroy.failed"
  fi
done
JENV_NAME="${JENV_NAME_SAVE}"

# Final Check
f_get_env_files
f_get_my_nova_instances
if [[ -n "$env_files" ]] || [[ -n "$my_instances" ]]; then
  echo " ! Juju env files:"
  echo "$env_files"
  echo " ! Instances:"
  echo "$my_instances"
  echo " ! juju1_bighammer_destroy failed.  Manual intervention required."
  echo "_:destroy.failed"

  # Never return if destroy failed; block to require human intervention.
  sleep infinity
fi
