#!/bin/bash -e
#
#  Common functions for UOSCI jobs, specific to OPENSTACK-INSTALLER functionality.

function f_osi_broom() {
# Clean-up for openstack-installer; remove .cloud-install dir and contents.

  [[ -z "$OSI_CLOUD_INSTALL_DIR" ]] && export OSI_CLOUD_INSTALL_DIR="$HOME/.cloud-install"

  if [[ -d $OSI_CLOUD_INSTALL_DIR ]]; then
    # Look for osi alternate juju environment, destroy if found
    OSI_JUJU_HOME="${OSI_CLOUD_INSTALL_DIR}/juju"
    if [[ -d $OSI_JUJU_HOME ]]; then
      export JUJU_HOME=$OSI_JUJU_HOME
      export JENV_NAME_TMP="$(juju switch || true)"
      if [[ -n $JENV_NAME_TMP ]]; then
        echo " - Blindly destroying juju enviro: $JENV_NAME_TMP ($JUJU_HOME)..."
        juju destroy-environment $JENV_NAME_TMP --force -y || true
      else
        echo " ! OSI juju home ($OSI_JUJU_HOME) exists, but could not determine osi's juju environment name."
      fi
      unset JUJU_HOME
      unset JENV_NAME_TMP
    fi

    echo " ! .cloud-install dir exists ($OSI_CLOUD_INSTALL_DIR), removing..."
    if [[ "$OSI_CLOUD_INSTALL_DIR" != *cloud-install* ]]; then
      (echo " ! OSI_CLOUD_INSTALL_DIR env var does not appear to be a cloud-install dir.  Bailing for safety."; exit 1)
    else
      # TODO: re-assess, re-enable some how
      echo " ! NOT IMPLEMENTED (deprecated, no sudo privileges)"
      exit 1
      #sudo rm -rf $OSI_CLOUD_INSTALL_DIR
    fi
  fi
}


function f_osi_config_file_pull_and_sanitize() {
# Pull openstack-installer config yaml to workspace (sanitize keys and passwords)
  DEST_FILE="$WORKSPACE/osi-config.yaml"
  echo " . Copying (sanitized) $OSI_CONFIG_FILE to $DEST_FILE"
  sed -e 's/\(key:\).*$/\1/' \
      -e 's/\(password:\).*$/\1/' \
      -e 's/\(email:\).*$/\1/' \
      $OSI_CONFIG_FILE > $DEST_FILE
}


function f_osi_determine_config_file() {
# Switch to determine config file name.
  case "$OSI_MODE" in
    "multi"|"single"|"autopilot")
      export OSI_CONFIG_FILE="${JENKINS_HOME}/osi-config-${OSI_MODE}.yaml"
      ;;
    *)
      # TO-DO:  Allow use of an arbitrary config yaml file, passed from Jenkins workspace.
      echo " ! Unknown OSI_MODE value while determining openstack-installer config file name."
      exit 1
      ;;
  esac

  if [[ -f "$OSI_CONFIG_FILE" ]]; then
    f_osi_config_file_pull_and_sanitize
  else
    echo " ! OSI_CONFIG_FILE (${OSI_CONFIG_FILE}) not found."
    exit 1
  fi
}


function f_osi_deploy() {
  f_line "OSI deploy"
  f_osi_determine_config_file

  # Generally takes < 1hr, but have had some infinite deploy occasions, ie:
  # https://github.com/Ubuntu-Solutions-Engineering/openstack-installer/issues/839
  # Hence, force a timeout.
  timeout 7200 openstack-install --debug --headless \
      --openstack-release $OPENSTACK_RELEASE --series $UBUNTU_RELEASE \
      --constraints "$UOSCI_JUJU_CONSTRAINTS $UOSCI_MAAS_TAGS" \
      -c $OSI_CONFIG_FILE |& tee $WORKSPACE/osi-deploy.txt
  ret_code=${PIPESTATUS[0]}

  if [[ $ret_code == "0" ]]; then
    echo " . OK:  openstack-install (deploy)"
    touch $WORKSPACE/fyi-${test:0:4}.osi-deploy.ok
  else
    echo " ! Failed:  openstack-install (deploy)"
    touch $WORKSPACE/fyi-${test:0:4}.osi-deploy.failed
  fi
}


function f_osi_collect() {
# Collect openstack-installer logs and files as tarball artifact
  f_line "OSI artifact collection"

  if [[ -n $OSI_CLOUD_INSTALL_DIR ]] && \
     [[ -d $OSI_CLOUD_INSTALL_DIR ]]; then
    echo " . Archiving $OSI_CLOUD_INSTALL_DIR as artifact..."
    tar -cjf $WORKSPACE/cloud-install.tar.bz2 $OSI_CLOUD_INSTALL_DIR \
        --exclude "$OSI_CLOUD_INSTALL_DIR/juju" \
        --exclude "$OSI_CLOUD_INSTALL_DIR/juju?" \
        --exclude "$OSI_CLOUD_INSTALL_DIR/ssh" \
        --exclude "$OSI_CLOUD_INSTALL_DIR/local-charms"
    ls -alh $WORKSPACE/cloud-install.tar.bz2
  else
    echo " ! Failed to archive $OSI_CLOUD_INSTALL_DIR as artifact."
    echo " ! OSI cloud installer dir not found @ $OSI_CLOUD_INSTALL_DIR."
    touch $WORKSPACE/fyi-${test:0:4}.cloud-install-dir-check.failed
  fi
}


function f_osi_uninstall() {
# Note: does not uninstall openstack-installer packages. Uninstalls (destroys)
# the cloud which was deployed by openstack-installer.

  if [[ "${NO_POST_DESTROY^^}" != "TRUE" ]]; then
    # osi uninstall has no unattended feature
    # https://github.com/Ubuntu-Solutions-Engineering/openstack-installer/issues/700
    # Command may be something like:
    # sudo openstack-installer --uninstall --headless -y
    # echo " ! Not-Implemented:  openstack-installer --uninstall (requires interactive)"
    #
    # Work-around:

    echo 'y' | openstack-install -u
    ret_code=${PIPESTATUS[1]}

    if [[ $ret_code == "0" ]]; then
     echo " . OK:  openstack-install --uninstall."
      touch $WORKSPACE/fyi-${test:0:4}.osi-uninstall.ok
    else
      echo " ! Failed:  openstack-install --uninstall."
      touch $WORKSPACE/fyi-${test:0:4}.osi-uninstall.failed
      juju destroy-environment maas -y || :
    fi
  fi
}


function f_osi_basic_deploy_cli_check() {
# For the default OSI deployment, perform basic command line checks against the cloud
  echo "Confirming basic CLI functionality and authentication..."
  . $OPENSTACK_CHARM_TESTING_ROOT/novarc
  tmp_cmds="$(mktemp)"
  cat > "$tmp_cmds" << "--EOF"
  glance image-list
  keystone ec2-credentials-list
  keystone endpoint-list
  keystone role-list
  keystone service-list
  keystone tenant-list
  keystone token-get
  keystone user-list
  keystone user-role-list
  neutron net-list
  neutron router-list
  neutron subnet-list
  nova flavor-list
  nova floating-ip-list
  nova floating-ip-pool-list
  nova keypair-list
  nova host-list
  nova hypervisor-list
  nova image-list
  nova network-list
  nova secgroup-list
  nova service-list
--EOF
  while read cmd; do
    echo -n "${cmd}: "
    $cmd &> /dev/null && echo 'ok' ${PIPESTATUS[0]} || { echo 'cli-fail' ${PIPESTATUS[0]}; touch $WORKSPACE/fyi-${test:0:4}.cli.failed; }
  done < $tmp_cmds
  rm -fv $tmp_cmds
}


function f_osi_basic_deploy_validation() {
# Basic sanity checks of the deployed cloud
  f_line "Basic OSI cloud validation"

  # Check glance image id
  img=$(glance image-list | grep ubuntu | grep active | awk '{ print $2 }')
  if echo $img | grep -o "[0-9a-fA-F]\{8\}-[0-9a-fA-F]\{4\}-[0-9a-fA-F]\{4\}-[0-9a-fA-F]\{4\}-[0-9a-fA-F]\{12\}"; then
    echo " . Glance image ID check OK: ${img}"
    touch $WORKSPACE/fyi-${test:0:4}.osi-glance-image-uuid.ok
  else
    echo " ! Glance image ID not a UUID: ${img}"
    touch $WORKSPACE/fyi-${test:0:4}.osi-glance-image-uuid.failed
    glance image-list
  fi

  # Check keystone users
  if [[ $(keystone user-list | egrep "admin|glance|image-stream|nova|ubuntu" | wc -l) == "5" ]]; then
    echo " . Keystone users check OK."
    touch $WORKSPACE/fyi-${test:0:4}.osi-keystone-users.ok
  else
    echo " ! Keystone users check failed."
    touch $WORKSPACE/fyi-${test:0:4}.osi-keystone-users.failed
    keystone user-list
  fi

  # Check keystone endpoints
  if [[ $(keystone endpoint-list | egrep ":5000|:9292|:8774|:9696" | grep RegionOne | wc -l) == "4" ]]; then
    echo " . Keystone endpoints check OK."
    touch $WORKSPACE/fyi-${test:0:4}.osi-keystone-endpoints.ok
  else
    echo " ! Keystone endpoints check failed."
    touch $WORKSPACE/fyi-${test:0:4}.osi-keystone-endpoints.failed
    keystone endpoint-list
  fi

  # Check nova services
  if [[ $(nova service-list | egrep "nova-conductor|nova-cert|nova-scheduler|nova-compute" | egrep "enabled.*up" | wc -l) == "4" ]]; then
    echo " . Nova services list check OK."
    touch $WORKSPACE/fyi-${test:0:4}.osi-nova-services.ok
  else
    echo " ! Nova services list check failed."
    touch $WORKSPACE/fyi-${test:0:4}.osi-nova-services.failed
    nova service-list
  fi
}


function f_osi_refresh() {
# Git the openstack-installer repo
  [[ -z "$OSI_REPO" ]] && export OSI_REPO="https://github.com/Ubuntu-Solutions-Engineering/openstack-installer"
  [[ -z "$OSI_ROOT" ]] && export OSI_ROOT="$HOME/tools/openstack-installer"

  if [[ -d $OSI_ROOT ]]; then
    echo " - Removing existing dir @ $OSI_ROOT"
    rm -rf $OSI_ROOT
  fi

  echo " + Creating new shallow clone of $OSI_REPO @ $OSI_ROOT."
  git clone -q --depth=1 $OSI_REPO $OSI_ROOT
  git --git-dir=$OSI_ROOT/.git --work-tree=$OSI_ROOT log --pretty=fuller
  git --git-dir=$OSI_ROOT/.git --work-tree=$OSI_ROOT remote show origin
}


function f_osi_built_in_ci_tests() {
# Call OSI built-in tests
  f_line "Running Built-In OSI Tests"
  case "$OSI_MODE" in
    "multi"|"single"|"autopilot")
      echo " . Starting OSI built-in ${OSI_MODE} mode test..."
      f_osi_refresh
      cd $OSI_ROOT/integration
      py.test-3 tests/test_${OSI_MODE}.py |& tee $WORKSPACE/osi-test-${OSI_MODE}.txt
      ret_code=${PIPESTATUS[0]}
      ;;
    *)
      echo " ! Unknown OSI_MODE value for OSI built-in test determination: ${OSI_MODE}"
      ;;
  esac

  if [[ $ret_code == "0" ]]; then
    echo " . OK:  OSI built-in tests"
    touch $WORKSPACE/fyi-${test:0:4}.osi-tests.ok
  else
    echo " ! Failed:  OSI built-in tests"
    touch $WORKSPACE/fyi-${test:0:4}.osi-tests.failed
  fi
}
