#!/bin/bash -e
#
#  Common functions for UOSCI jobs, specific to GERRIT functionality.

# Move to sim
#[[ -z "$GERRIT_CHANGE_NUMBER" ]] && export GERRIT_CHANGE_NUMBER="-1"
#[[ -z "$GERRIT_TOPIC" ]] && export GERRIT_TOPIC="simulated topic"
#[[ -z "$GERRIT_PROJECT" ]] && export GERRIT_PROJECT="nosuch/project"
#[[ -z "$GERRIT_CHANGE_ID" ]] && export GERRIT_CHANGE_ID="abbaacaddabbafabbaaaa123456c0ffeedeadbeef"
#[[ -z "$GERRIT_PATCHSET_NUMBER" ]] && export GERRIT_PATCHSET_NUMBER="-100"
#[[ -z "$GERRIT_PATCHSET_UPLOADER_NAME" ]] && export GERRIT_PATCHSET_UPLOADER_NAME="John Doe"
#[[ -z "$GERRIT_PATCHSET_REVISION" ]] && export GERRIT_PATCHSET_REVISION="boot5fa7ca70000000abcdef0000000000000001"

#[[ -z "$BUILD_ID" ]] && export BUILD_ID="sim-$(uuidgen)"
#[[ -z "$JOB_NAME" ]] && export JOB_NAME="simulation-pipeline"
#[[ -z "$UOSCI_GERRIT_VERDICT" ]] && export UOSCI_GERRIT_VERDICT="UNKNOWN"


function f_gerrit_set_vote_values(){
    case "$UOSCI_GERRIT_VERDICT" in
      "SUCCESS")
        export UOSCI_GERRIT_VOTE_VERIFIED="1"
        export UOSCI_GERRIT_VOTE_CRVW="0"
        ;;
      "FAILURE")
        export UOSCI_GERRIT_VOTE_VERIFIED="-1"
        export UOSCI_GERRIT_VOTE_CRVW="0"
        ;;
      "UNSTABLE")
        export UOSCI_GERRIT_VOTE_VERIFIED="0"
        export UOSCI_GERRIT_VOTE_CRVW="0"
        ;;
      "UNKNOWN")
        export UOSCI_GERRIT_VOTE_VERIFIED="0"
        export UOSCI_GERRIT_VOTE_CRVW="0"
        ;;
      *)
        echo " ! Unknown UOSCI_GERRIT_VERDICT value."
        touch $WORKSPACE/uosci-gerrit-verdict.failed
    esac
}


function f_check_out_universal(){
# Handle all use cases by all jobs to check out or clone a repo.
  f_line $FUNCNAME

  # No need to check out at all if working with a layerci built asset
  if f_layerci_in_effect; then
    f_set_base_name_from_gerrit_project
    echo " . Not checking out repo for $BASE_NAME (${BRANCH-$GERRIT_PROJECT}).  Layer CI is in effect."
    return 0
  fi

  # Detect repo approach and fetch/clone/branch the asset
  if [[ ! -z "$GERRIT_REFSPEC" ]] && [[ "$GERRIT_CHANGE_NUMBER" != "-1" ]]; then
    f_gerrit_git_fetch_refspec
  elif [[ ! -z "$BASE_NAME" ]] && [[ -z "$BRANCH" ]]; then
    f_git_clone_by_base_name
  elif [[ ! -z "$BASE_NAME" ]] && [[ ! -z "$BRANCH" ]]; then
    f_check_out_lp_branch
  else
    echo " ! Unable to determine repo retrieval method.  Missing env vars?"
    exit 1
  fi
}


function f_git_clone_by_base_name(){
# Git clone based on BASE_NAME and optional GIT_BRANCH (default to master)
  f_line $FUNCNAME

  [[ -z "$GIT_BRANCH" ]] && export GIT_BRANCH="master"

  if [[ -z "$BASE_NAME" ]]; then
    echo " ! Missing the required GIT env vars."
    exit 1
  fi

  f_set_co_dir
  export GERRIT_GIT_HOST="github.com"
  export GERRIT_PROJECT="openstack/charm-${BASE_NAME}"
  # ex. https://github.com/openstack/charm-swift-proxy

  if [[ -d $CO_DIR ]]; then
    echo " - Removing existing checkout @ $CO_DIR"
    rm -rf ${CO_DIR:?}
  fi

  mkdir -p $CO_DIR
  echo " + Cloning https://${GERRIT_GIT_HOST}/${GERRIT_PROJECT} --> $CO_DIR"
  git clone https://${GERRIT_GIT_HOST}/${GERRIT_PROJECT} $CO_DIR -b $GIT_BRANCH --depth 1
  f_git_announce_info
}


function f_set_base_name_from_gerrit_project(){
  if [[ -z "$GERRIT_PROJECT" ]]; then
    echo " ! Missing the required GERRIT_PROJECT env var."
    exit 1
  fi
  if [[ "$GERRIT_PROJECT" == *charm* ]]; then
    # ex. openstack/charm-ceilometer
    #     openstack/charm-openstack-dashboard
    export BASE_NAME="$(echo $GERRIT_PROJECT | cut -f 2 -d "/" | cut -f 2,3,4,5 -d "-")"
  else
    # ex. openstack-infra/project-config
    export BASE_NAME=$GERRIT_PROJECT
  fi
}


function f_gerrit_git_fetch_refspec(){
# Shallow clone a project from gerrit and apply the proposed change from a refspec value
  f_line $FUNCNAME

  export GERRIT_GIT_HOST="review.openstack.org"
  #export GERRIT_GIT_HOST="github.com"

  if [[ -z "$GERRIT_PROJECT" ]] || [[ -z "$GERRIT_REFSPEC" ]] || [[ -z "$GERRIT_BRANCH" ]]; then
    echo " ! Missing the required GERRIT env vars."
    exit 1
  fi

  f_set_base_name_from_gerrit_project
  f_set_co_dir

  echo "BASE_NAME: ${BASE_NAME}"
  echo "CO_DIR: ${CO_DIR}"

  if [[ -d $CO_DIR ]]; then
    echo " - Removing existing checkout @ $CO_DIR"
    rm -rf ${CO_DIR:?}
  fi

  mkdir -p $CO_DIR
  echo " + Cloning https://${GERRIT_GIT_HOST}/${GERRIT_PROJECT} --> $CO_DIR"
  git clone https://${GERRIT_GIT_HOST}/${GERRIT_PROJECT} $CO_DIR -b $GERRIT_BRANCH --depth 1
  cd $CO_DIR
  echo " + Fetching ${GERRIT_REFSPEC}"
  git fetch https://${GERRIT_GIT_HOST}/${GERRIT_PROJECT} ${GERRIT_REFSPEC} --depth 1
  echo " + Checking out FETCH_HEAD"
  git checkout FETCH_HEAD -q
  f_git_announce_info
}


function f_git_announce_info(){
  cd $CO_DIR
  echo " . git branch:"
  git branch -av
  echo " . git reflog:"
  git reflog
  echo " . size on disk:"
  du -hsc $CO_DIR
}


function f_gerrit_set_blob_vars(){
  export UOSCI_BUCKET_NAME="uosci-artifacts"
  export UOSCI_BLOB_STORE="https://openstack-ci-reports.ubuntu.com/artifacts"  # no trailing slash!

  if [[ ! -z "$GERRIT_CHANGE_NUMBER" ]]; then
    # GERRIT CHANGE TRIGGER
    export UOSCI_BLOB_PATH="${JOB_NAME}/${GERRIT_PROJECT}/${GERRIT_CHANGE_NUMBER}/${GERRIT_PATCHSET_NUMBER}/${BUILD_ID}"
  else
    # BASE NAME TRIGGER
    export UOSCI_BLOB_PATH="${JOB_NAME}/openstack/charm-${BASE_NAME}/${BUILD_ID}"
  fi
  export UOSCI_BLOB_URL="${UOSCI_BLOB_STORE}/${UOSCI_BLOB_PATH}/index.html"

  # Temp Debug
  echo "UOSCI_BLOB_URL:  ${UOSCI_BLOB_URL}"
}


function f_upload_this_build_archive_to_bucket(){
  f_line "Create index file and upload artifacts to bucket"
  echo "${WORKSPACE} --> $UOSCI_BLOB_PATH"

  f_gerrit_create_artifact_index
  echo "Uploading..."
  f_novarc_artifactcloud
  timeout 600 swift upload $UOSCI_BUCKET_NAME --object-name $UOSCI_BLOB_PATH ${WORKSPACE}

  # Set content-type for txt and html files
  export OBJECT_MATCH_PATTERN="$UOSCI_BLOB_PATH"
  f_bucket_reset_txt_html_object_headers

  # Temp Debug
  echo "UOSCI_BLOB_URL:  ${UOSCI_BLOB_URL}"
}


function f_bucket_upload_index_of_indices(){
  f_line "Upload index of indices to blob bucket"
  # The Liberty swift client does not set headers on upload (with radosgw
  # object storage), hence the post after upload.
  set -x
  timeout 600 swift upload -c -H "content-type:text/html" $UOSCI_BUCKET_NAME index.html
  timeout 600 swift post -H "content-type:text/html" $UOSCI_BUCKET_NAME index.html
  timeout 600 swift stat $UOSCI_BUCKET_NAME index.html
  set +x
}


function f_gerrit_set_link_url_vars(){
  export UOSCI_WIKI_LINK="https://wiki.openstack.org/wiki/ThirdPartySystems/Canonical_Charm_CI"
  export UOSCI_COMMUNITY_README_LINK="http://docs.openstack.org/developer/charm-guide/"
}


function f_gerrit_get_verdicts(){
  f_gerrit_set_blob_vars
  export UOSCI_GERRIT_STEP_VERDICTS="$(env | grep -oP "TRIGGERED_BUILD_RESULT_\K.*" | grep -v "_RUN_" | sed "s#=# $UOSCI_BLOB_URL : #g" | sed -e 's/^/* /')" ||:
  export UOSCI_GERRIT_STEP_VERDICTS_RAW="$(env | grep -oP "TRIGGERED_BUILD_RESULT_\K.*" | grep "_RUN_" | sed "s#=#:#g" | sort)" ||:
}


function f_gerrit_construct_message(){
  export _MSG_FINAL="Build ${UOSCI_GERRIT_VERDICT} by Canonical OpenStack Charm CI.
$UOSCI_GERRIT_STEP_VERDICTS

Re-test comment options:

- 'recheck' (run/re-run the default upstream and charm test sets)
- 'charm-recheck' (run/re-run the default charm verify test set)
- 'charm-recheck-full' (run/re-run the full charm amulet test set - may take hours!)

For more info, see ${UOSCI_COMMUNITY_README_LINK} and ${UOSCI_WIKI_LINK}."

  echo "$_MSG_FINAL" > $WORKSPACE/gerrit-review-message.txt
  echo "$UOSCI_GERRIT_STEP_VERDICTS_RAW" > $WORKSPACE/gerrit-step-verdicts-raw.txt
  cat $WORKSPACE/gerrit-step-verdicts-raw.txt
}


function f_gerrit_review_it(){
  f_gerrit_set_blob_vars
  f_gerrit_set_link_url_vars
  f_gerrit_set_vote_values
  env | grep GERRIT | sort
  f_gerrit_construct_message

  if [[ "$GERRIT_CHANGE_NUMBER" == "-1" ]]; then
    # SIMULATED BUILD (LIKELY A SANDBOX)
    f_line "Post a gerrit review"
    echo " . SIMULATION ONLY.  WOULD HAVE DONE:"
    echo ssh -p 29418 uosci-testing-bot@review.openstack.org gerrit review $GERRIT_CHANGE_NUMBER,$GERRIT_PATCHSET_NUMBER '--message='"\"$_MSG_FINAL\"" --verified $UOSCI_GERRIT_VOTE_VERIFIED --code-review $UOSCI_GERRIT_VOTE_CRVW
  elif [[ "$GERRIT_CHANGE_NUMBER" != "-1" ]] && [[ ! -z "$GERRIT_CHANGE_NUMBER" ]]; then
    # THE REAL DEAL - POST A REVIEW
    f_line "Post a gerrit review"
    set -x
    ssh -p 29418 uosci-testing-bot@review.openstack.org gerrit review $GERRIT_CHANGE_NUMBER,$GERRIT_PATCHSET_NUMBER '--message='"\"$_MSG_FINAL\"" --verified $UOSCI_GERRIT_VOTE_VERIFIED --code-review $UOSCI_GERRIT_VOTE_CRVW
    set +x
  else
    # DOES NOT APPEAR TO BE GERRIT-CHANGE-TRIGGERED
    echo " . NOT A GERRIT REVIEW"
  fi
}


function f_bucket_create_index_of_indices(){
  TITLE="Ubuntu OpenStack Charm CI - Index of Recent Test Results"
  OUT_FILE="$WORKSPACE/index.html"
  echo "" > $OUT_FILE

  f_line "Create index of indices ($OUT_FILE)"


      add_line "<html><head><title>"
      add_line "$TITLE"
      add_line "</title>"
      add_line "<meta http-equiv=\"content-type\" content=\"text/html; charset=UTF-8\">"
      add_line "<link rel=\"stylesheet\" type=\"text/css\" media=\"all\" href=\"https://assets.ubuntu.com/sites/guidelines/css/latest/ubuntu-styles.css\"/>"
      add_line "<link rel=\"icon\" type=\"image/png\" href=\"static/uosci-index-16-transparent.png\"/>"
      add_line "</head><body>"
      add_line "<div class=\"wrapper\"><div id=\"main-content\" class=\"inner-wrapper\"><div class=\"row no-border\">"
      add_line "  <H3>$TITLE</H3>"
      add_line "  <table>"

  # Get index string from key 4, reverse sort by key 2 (date)
  INDICES=$(timeout 600 swift list uosci-artifacts -l | grep -e 'index.html$' | grep -v ' index.html' | sort -rk 2 | awk '{ print $4 }')

  for i in $INDICES; do
    # link="<a href=\"${i}\">VIEW</a>"
    # row="<tr><td>$(echo $i | sed -e 's#/#</td><td>#g')</td</tr>"
    # row_linked="$(echo $row | sed -e "s#index.html#${link}#")"
    # add_line "  $row_linked"

    link="<a href=\"${i}\">${i%"/index.html"}</a>"
    row="<tr><td style=\"text-align:left;\">$link</td</tr>"
      add_line "  $row"
  done
      add_line "  </table>"
      add_line "</div></div></div>"
      add_line "  <footer class=\"global clearfix\">"
      add_line "    <center>Generated $(date) by uosci-testing-bot.</center>"
      add_line "  </footer>"
      add_line "</body></html>"
}


function f_gerrit_create_artifact_index(){
  cd $WORKSPACE
  TITLE="UOSCI Test Results: #${BUILD_NUMBER} ${GERRIT_BRANCH} ${GERRIT_PROJECT} ${GERRIT_CHANGE_NUMBER} ${GERRIT_PATCHSET_NUMBER} ${GERRIT_PATCHSET_UPLOADER_NAME} ${BASE_NAME} ${GIT_BRANCH}"
  OUT_FILE="$WORKSPACE/index.html"
  echo "" > $OUT_FILE

      add_line "<html><head><title>"
      add_line "$TITLE"
      add_line "</title>"
      add_line "<meta http-equiv=\"content-type\" content=\"text/html; charset=UTF-8\">"
      add_line "<link rel=\"stylesheet\" type=\"text/css\" media=\"all\" href=\"https://assets.ubuntu.com/sites/guidelines/css/latest/ubuntu-styles.css\"/>"
      add_line "<link rel=\"icon\" type=\"image/png\" href=\"https://raw.githubusercontent.com/ryan-beisner/static-assets/master/uosci-bot-block-16-transparent.png\"/>"
      add_line "</head><body>"
      add_line "<div class=\"wrapper\"><div id=\"main-content\" class=\"inner-wrapper\"><div class=\"row no-border\">"
      add_line "  <H2>Canonical OpenStack Charm CI</H2>"
      add_line "  <H3>Test #${BUILD_NUMBER} ${GERRIT_PROJECT} ${GERRIT_CHANGE_NUMBER},${GERRIT_PATCHSET_NUMBER} for ${GERRIT_PATCHSET_UPLOADER_NAME}</H3>"
      add_line "  <hr/>"
      add_line "  <p><a href=\"$GERRIT_CHANGE_URL\">$GERRIT_CHANGE_URL</a></p>"
      add_line "  <pre style=\"font-size: 14px\">"
      add_line "Uploader: $GERRIT_PATCHSET_UPLOADER_NAME"
      add_line "Subject:  $GERRIT_CHANGE_SUBJECT"
      add_line "Project:  $GERRIT_PROJECT"
      add_line "Branch:   $GERRIT_BRANCH $GIT_BRANCH"
      add_line "Change:   $GERRIT_CHANGE_NUMBER  $GERRIT_PATCHSET_NUMBER"
      add_line "Test Set: $JOB_NAME (#$BUILD_NUMBER)"
      add_line "Build:    $BUILD_ID"
      add_line "  </pre>"
      add_line "  <hr/>"
      add_line "  <H3>Results</H3>"
      add_line "  <pre style=\"font-size: 14px\">"
      add_line "$(cat $WORKSPACE/gerrit-step-verdicts-raw.txt)"
      add_line "  </pre>"
      add_line "  <hr/>"
      add_line "  <H3>Build Artifacts</H3>"
      add_line "  <H4>${JOB_NAME}_${BUILD_NUMBER}</H4>"
      add_line "  <UL>"

  ROOT_FILES="$(find . -maxdepth 1 -mindepth 1 -type f | sort)"
  for FILE in $ROOT_FILES; do
      add_line "    <LI class=\"no-bullets\"><a href=\"$FILE\">$(basename $FILE)</a></LI>"
  done
      add_line "  </UL>"

  DIRS="$(find . -maxdepth 1 -mindepth 1 -type d | sort)"
  for DIRECTORY in $DIRS; do
      add_line "  <H4>$(basename $DIRECTORY)</H4>"
      add_line "  <UL>"
    for FILE in $(find $DIRECTORY -maxdepth 1 -mindepth 1 -type f | sort); do
      add_line "    <LI class=\"no-bullets\"><a href=\"$FILE\">$(basename $FILE)</a></LI>"
    done

    # Handle only 1 level of subdirectory
    DIRS_INNER="$(find $DIRECTORY -maxdepth 1 -mindepth 1 -type d | sort)"
    for DIRECTORY_INNER in $DIRS_INNER; do
      add_line "      <UL>"
      add_line "    <LI class=\"no-bullets\">$(basename $DIRECTORY_INNER):</LI>"
      add_line "    <UL>"
        for FILE_INNER in $(find $DIRECTORY_INNER -maxdepth 1 -mindepth 1 -type f | sort); do
      add_line "      <LI class=\"no-bullets smaller\"><a href=\"$FILE_INNER\">$(basename $FILE_INNER)</a></LI>"
        done
      add_line "    </UL>"
      add_line "      </UL>"
    done

      add_line "  </UL>"
  done
      add_line "  <hr/>"
      add_line "  <H4>See also:</H4>"
      add_line "  <UL>"
      add_line "    <LI class=\"no-bullets\"><a href=\"${UOSCI_COMMUNITY_README_LINK}\">Developing OpenStack Charms</a></LI>"
      add_line "    <LI class=\"no-bullets\"><a href=\"${UOSCI_WIKI_LINK}\">Canonical OpenStack Charm CI Third Party Wiki</a></LI>"
      add_line "  </UL>"
      add_line "</div></div></div>"
      add_line "  <footer class=\"global clearfix\">"
      add_line "    <center>Generated $(date) by uosci-testing-bot.</center>"
      add_line "  </footer>"
      add_line "</body></html>"

  stat $OUT_FILE ||:
}


function f_bucket_reset_txt_html_object_headers(){
  # This is a quick and dirty fix to set object header data in radosgw/swift container.
  # Objects uploaded with >= liberty clients are likely set as binary content-type,
  # which makes .txt and .html download instead of display in the user browser.

  [[ -z "$OBJECT_MATCH_PATTERN" ]] && export OBJECT_MATCH_PATTERN=""

  f_line "Updating *.txt and *.html content-type object headers"

  work_dir="$(mktemp -d)"
  tmp_object_list="$(mktemp)"

  echo " . Selecting .txt and .html files in $UOSCI_BUCKET_NAME..."
  echo " . OBJECT_MATCH_PATTERN:  $OBJECT_MATCH_PATTERN"
  timeout 600 swift list $UOSCI_BUCKET_NAME | egrep ".txt$|.html$" | grep "$OBJECT_MATCH_PATTERN" > $tmp_object_list ||:

  if [[ "$(wc -l $tmp_object_list)" == "0 $tmp_object_list" ]]; then
    echo " . No objects found to update"
  fi

  cd $work_dir

  while read -r line; do
    # Trim whitespace if it exists
    line="$(echo $line | sed 's/^[ \t]*//;s/[ \t]*$//')"

    # Update object
    if [[ -n "$line" ]]; then
      echo " - Updating object:  $line"
      if [[ "$line" == *.html ]]; then
        timeout 600 swift post $UOSCI_BUCKET_NAME $line -H "content-type:text/html"
      elif [[ "$line" == *.txt ]]; then
        timeout 600 swift post $UOSCI_BUCKET_NAME $line -H "content-type:text/plain"
      else
        echo " ! Ignoring:  $line"
      fi
    else
      # Let's not modify the container itself!
      echo " ! Skipping dangerous variable as read: $line"
    fi
  done < "$tmp_object_list"

  rm -fv $tmp_object_list

  f_line "END: Updating txt and html object headers"
}


function add_line(){
  echo "$1" >> $OUT_FILE
}


function f_gerrit_trigger_whitelist_check(){
    export GERRIT_ACCOUNT_WHITELIST="$UOSCI_BOT_CONTROL_ROOT/tactical/gerrit-trigger-whitelist.txt"
    $UOSCI_BOT_CONTROL_ROOT/tactical/gerrit-trigger-whitelist-check.sh
}
