#!/bin/bash -e
#
#  Common functions for UOSCI jobs
#
#  NOTE(beisner):  Many/most of these funcs rely on environment variables being
#  set by the build parameters or the calling script(s).

. ${OSCI_ROOT}/run/job-parts/osci_repos_common.sh
. ${OSCI_ROOT}/run/job-parts/osci_openstack_common.sh
. ${OSCI_ROOT}/run/job-parts/osci_mojo_common.sh
. ${OSCI_ROOT}/run/job-parts/osci_gerrit_common.sh
. ${OSCI_ROOT}/run/job-parts/osci_layerci_common.sh


function retry_command() {
# Work-around wrapper to deal with things like https://bugs.launchpad.net/juju/+bug/1709324
    command=$@
    i=0
    attempts=3
    while [ $i -lt $attempts ]; do
        $command && break
        let "i+=1"
        sleep 10
    done
}


function f_juju_wait() {
    f_line $FUNCNAME
    sleep 30
    [[ -z "$JUJU_WAIT_TIMEOUT" ]] && export JUJU_WAIT_TIMEOUT=1800
    timeout $JUJU_WAIT_TIMEOUT $JUJU_WAIT_ROOT/juju-wait -v ||\
        (
            echo " ! Model failed to settle within ${JUJU_WAIT_TIMEOUT}s"
            touch $WORKSPACE/fyi-${test:0:4}.${test}-juju-wait.failed
        )
}


function f_juju_wait_on_workload() {
    f_line $FUNCNAME
    sleep 30
    [[ -z "$JUJU_WAIT_TIMEOUT" ]] && export JUJU_WAIT_TIMEOUT=1800
    timeout $JUJU_WAIT_TIMEOUT $JUJU_WAIT_ROOT/juju-wait -vw ||\
        (
            echo " ! Model failed to settle within ${JUJU_WAIT_TIMEOUT}s"
            touch $WORKSPACE/fyi-${test:0:4}.${test}-juju-wait.failed
        )
}


function f_ssh_known_hosts_clean() {
  echo "Overwriting with a clean known_hosts file on ${NODE_NAME}..."
  cat $OSCI_ROOT/deploy/assets/launchpad_host_keys > /var/lib/jenkins/.ssh/known_hosts
  cat $OSCI_ROOT/deploy/assets/git_host_keys >> /var/lib/jenkins/.ssh/known_hosts
  cat $OSCI_ROOT/deploy/assets/gerrit_host_keys >> /var/lib/jenkins/.ssh/known_hosts
  cat $OSCI_ROOT/deploy/assets/other_host_keys >> /var/lib/jenkins/.ssh/known_hosts
  chmod 600 /var/lib/jenkins/.ssh/known_hosts
  chown jenkins:jenkins /var/lib/jenkins/.ssh/known_hosts
}


function f_workspace_cleanup() {
  if [[ -d "$WORKSPACE" ]]; then
    if [[ -n "$(echo $WORKSPACE | grep '/var/lib/jenkins')" ]]; then
      echo " - Removing contents of workspace ${WORKSPACE}"
      rm -rf ${WORKSPACE:?}/*
    fi
  fi
}


function f_fresh_clean_start() {
  f_line $FUNCNAME
  f_workspace_cleanup
  f_ssh_known_hosts_clean
  f_co_cleanup
  if [[ "${SKIP_JUJU_FOO^^}" != "TRUE" ]]; then
      f_tool_refresh_cti
      f_tool_refresh_juju_wait
      f_juju_destroy_model
  fi
}


function f_touch_try() {
# Create touchfile, indicating that we tried to run this test.
# Valid only in the LP workflow.
  if [[ -n "$MP_TRIGGER" ]]; then
    touch $WORKSPACE/try-${test:0:4}
  fi
}


function f_env_no_pass() {
# Dump environment variables, except for password vars
  env | egrep -v "PASSWORD|password|COOKIE|cookie|OS_|USER|user|PROXY|proxy" | \
    sort 2>&1 | tee $WORKSPACE/env.txt
}


function f_list_nova_instances() {
  # DEBUG: list all running instances
  f_novarc_undercloud
  if [[ "${JENV_NAME^^}" != *MAAS* ]]; then
    f_line "Listing nova instances to artifact"
    # Trap true, as instances may not exist
    nova list | grep "\-${JENV_NAME}\-machine" &> $WORKSPACE/nova-instances-$JENV_NAME.txt || :
    nova list &> $WORKSPACE/nova-instances-all.txt || :
  fi
}


function f_list_neutron_ports() {
  # DEBUG: list all tenant neutron ports
  f_novarc_undercloud
  if [[ "${JENV_NAME^^}" != *MAAS* ]]; then
    f_line "Listing neutron ports artifact"
    # Trap true, as ports may not exist
    neutron port-list &> $WORKSPACE/neutron-ports.txt || :
  fi
}


function f_delete_neutron_ext_net_ports() {
  # CLEANUP: delete neutron ext-net ports, commonly added by mojo or other
  # post-deployment configuration for the network node's 2nd nic.
  f_novarc_undercloud
  if [[ -z "${MODEL_NAME}" ]]; then
    echo " ! Missing required env var MODEL_NAME"
    exit 1
  fi
  if [[ ! -f "${UOSCI_BOT_CONTROL_ROOT}/tools/juju-sym-switch" ]]; then
    f_tool_refresh_bot_control
  fi
  if [[ "${JENV_NAME^^}" != *MAAS* ]]; then
    f_line "Deleting undercloud tenant's neutron ext-net ports if they exist"
    f_novarc_undercloud
    f_activate_botcontrol_tools_virtualenv
    echo "DELETING ext-ports:"
    ${UOSCI_BOT_CONTROL_ROOT}/tools/port-cleanup -pn "juju-.*-${MODEL_NAME}.*_ext-port" -ps "DOWN" --delete
  fi
}


function f_delete_secgroups() {
  # CLEANUP: delete nova security groups which are often not deleted
  # by Juju destroy.
  #   ex:   juju-osci-sv01-31
  #   ref:  https://bugs.launchpad.net/juju-core/+bug/1333162
  #         https://bugs.launchpad.net/juju/+bug/1570216

  f_novarc_undercloud
  if [[ -z "${JENV_NAME}" ]]; then
    echo " ! Missing required env var JENV_NAME"
    exit 1
  fi
  if [[ "${JENV_NAME^^}" != *MAAS* ]]; then
    f_line "Deleting undercloud tenant's nova sec groups for $JENV_NAME"
    for i in $(nova secgroup-list | egrep "juju-osci-sv|juju-osci-lb|juju-osci-tk" | grep "$JENV_NAME" | awk '{ print $2 }'); do echo $i; echo NOT DELETING: nova secgroup-delete $i; done
  fi
}



function f_controller_id() {
    export CONTROLLER_ID="$(juju controllers --format yaml | awk '/ uuid: /{ print $2 }')"
}


function f_model_id() {
    export MODEL_ID="$(juju show-model $(juju switch)  | awk '/ model-uuid: /{ print $2 }')"
}


function f_subnet_id() {
    export SUBNET_ID="$(openstack network show $NETWORK_ID | awk '/ subnets /{ print $4 }')"
}


function f_post_destroy_info_dump() {
    f_activate_clients
    f_novarc_undercloud
    f_subnet_id
    f_controller_id
    f_model_id

    FILTER="Name|$NETWORK_ID"
        [ -n "$SUBNET_ID" ] && { FILTER+="|$SUBNET_ID"; }
        [ -n "$CONTROLLER_ID" ] && { FILTER+="|$CONTROLLER_ID"; }
        [ -n "$MODEL_ID" ] && { FILTER+="|$MODEL_ID"; }

    openstack port list            | egrep "$FILTER" | tee $WORKSPACE/post-destroy-port-list.txt
    openstack network list         | egrep "$FILTER" | tee $WORKSPACE/post-destroy-network-list.txt
    openstack subnet list          | egrep "$FILTER" | tee $WORKSPACE/post-destroy-subnet-list.txt
    openstack security group list  | egrep "$FILTER" | tee $WORKSPACE/post-destroy-secgroup-list.txt
}


function f_nice_finish() {
# Generally run at the end of every build
  if [[ "${NO_POST_DESTROY^^}" != "TRUE" ]]; then
    f_tmp_cleanup
    f_co_cleanup
  fi
  if [[ "${NO_POST_DESTROY^^}" != "TRUE" ]] && [[ "${SKIP_JUJU_FOO^^}" != "TRUE" ]]; then
    f_line "Destroy"
    f_juju_destroy_model
    f_post_destroy_info_dump

    f_delete_neutron_ext_net_ports

    # XXX: Disable cleanup until determined if still necessary; needs updated if so.
    # f_line "Post-Destroy Additional Clean-up"
    # f_delete_secgroups ||:
    # f_mojo_broom
    # f_osi_broom
  fi
}


function f_check_fail() {
# Check various indicators of test failure and force job failure if found.

  # Check bundletester report file for failure indicator
  if [[ -f ${BUILT_ASSET_DIR-$CO_DIR}/func-results.json ]]; then
    func_nonzero_retcodes=$(grep "\"\?returncode\"\?:" func-results.json | awk '{ print $2 }' | tr -d ',' | grep -v "0" ||:)
    # Ex. sample data from func-results.json:
    #      "returncode": 0,
    #      "returncode": 4,
    #      "returncode": 1,
    # Produces non-zero return codes:
    # 1
    # 4
    if [[ -n "$func_nonzero_retcodes" ]]; then
      echo " ! Bundletester report indicates one or more tests failed (ret code(s): $func_nonzero_retcodes)"
      echo " ! Exiting 1."
      exit 1
    else
      echo " . Bundletester report indicates no test failures (OK)"
    fi
  fi

  # Check for failure touchfiles and fail if found
  lso="$(ls $WORKSPACE -1)"
  if [[ "$lso" == *.failed* ]] || \
     [[ "$lso" == *.missing* ]] || \
     [[ "$lso" == *.bad* ]]; then
    echo " ! Exiting 1."
    exit 1
  fi
}


function f_post_mp_comment() {
# Post mp comment if triggered by a merge proposal
  if [[ -n "$MP_TRIGGER" ]]; then
    cd $OSCI_ROOT
    tools/mp_comment.py ||:
  fi
}


function f_set_build_descr() {
# Collect the FYI files for build description
  cd $WORKSPACE
  b_d="$(ls -1 fyi*.* 2> /dev/null | cut -d '.' -f 2,3 | tr '\n' ' ' ||:)"

  if [[ -n "$b_d" ]]; then
    # Output build description for jenkins plugin pickup by regex
    echo "_:${b_d}"
  fi
}


function f_get_func_results_file() {
# Copy functional test results file if it exists
  cp -fv ${BUILT_ASSET_DIR-$CO_DIR}/func-results.json $WORKSPACE/func-results.json.txt 2>/dev/null ||:
}


function f_set_co_dir() {
# Set the checkout dir
  mkdir -p $JENKINS_HOME/checkout/$EXECUTOR_NUMBER
  export CO_DIR=$JENKINS_HOME/checkout/$EXECUTOR_NUMBER/$BASE_NAME
}


function f_check_out_lp_branch() {
  f_line $FUNCNAME
  echo " ! lp branch checkout is now NotImplemented. If you hit this, raise a bug."
  exit 1
  # if [[ -z "$BASE_NAME" ]] || [[ -z "$BRANCH" ]]; then
  #   echo " ! No BRANCH and/or BASE_NAME env vars."
  #   touch $WORKSPACE/fyi-${test:0:4}.env-vars.bad
  # else
  #  mkdir -pv $JENKINS_HOME/checkout
  #  f_set_co_dir
  #  MAKEFILE=$CO_DIR/Makefile

  #  if [[ -d $CO_DIR ]]; then
  #    echo " - Removing existing checkout @ $CO_DIR"
  #    rm -rf ${CO_DIR:?}
  #  fi
  #  echo " + Creating new checkout of $BRANCH @ $CO_DIR."
  #  bzr co --lightweight $BRANCH $CO_DIR
  #  bzr revno $CO_DIR 2>&1 | tee $WORKSPACE/branch-info.txt
  #  bzr info $CO_DIR 2>&1 | tee -a $WORKSPACE/branch-info.txt
  #  bzr log $CO_DIR -l1 2>&1 | tee -a $WORKSPACE/branch-info.txt
  # fi

  # if [[ "${CHARMHELPER_SYNC^^}" == "TRUE" ]]; then
  #  # sync charmhelpers if boolean is set
  #  make sync
  # fi
}


function f_run_make_or_tox_test() {
# Inspect Makefile and tox.ini for matches and run the first match.
  cd ${BUILT_ASSET_DIR-$CO_DIR}

  # Sniff tox.ini and Makefile
  unset OSCI_TOX_TARGET_FOUND

  # XXX: deprecating makefile support
  # unset OSCI_MAKE_TARGET_FOUND
  # for t in "${sstring[@]}"; do
  #   if ${OSCI_ROOT}/tools/make_exec.py Makefile "$t" -X &> /dev/null; then
  #     echo " . Make target containing (\"$t\") exists"
  #     export OSCI_MAKE_TARGET_FOUND="True"
  #     break
  #   fi
  # done

  for t in "${tox_e_try[@]}"; do
    if grep "^\[testenv:${t}\]" tox.ini &> /dev/null; then
      echo " . Tox target (\"$t\") exists"
      export OSCI_TOX_TARGET_FOUND="$t"
      break
    fi
  done

  if [[ -n "$OSCI_MAKE_TARGET_FOUND" ]]; then
    echo " ! DEPRECATED: makefile runners"
    exit 1
  elif [[ -n "$OSCI_TOX_TARGET_FOUND" ]]; then
    echo " . Executing tox environment: $t"
    f_run_tox_environment $t
  else
    echo " ! Unable to determine make or tox targets to run"
    # echo "Tried to find make targets containing one of: ${sstring[*]}"
    echo "Tried to find tox a environment named: ${tox_e_try[*]}"
    touch $WORKSPACE/makefile-or-toxini.missing
  fi
}


function f_run_tox_environment() {
# Run tox with optional environment (target) as the parameter
  if [[ -n "$1" ]]; then
    PARAM1="-e $1"
  else
    PARAM1=""
  fi

  tox $PARAM1 2>&1 | tee $WORKSPACE/tox-output.txt; ret_code=${PIPESTATUS[0]}
  echo "Tox returned:  ${ret_code}"
  if [[ "${ret_code}" != "0" ]]; then
      touch $WORKSPACE/tox.failed
  fi
}


function f_proxy_if_proxy() {
# Set http proxy if already set for amulet
  if [[ -n "$AMULET_HTTP_PROXY" ]]; then
    export NO_PROXY=$OSCI_NO_PROXY
    export HTTP_PROXY=$AMULET_HTTP_PROXY
    export HTTPS_PROXY=$(echo $AMULET_HTTP_PROXY | sed 's/http/https/g')
    export http_proxy=$HTTP_PROXY
    export https_proxy=$HTTPS_PROXY
    echo " + Enabling HTTP(S) proxies."
  fi
}


function f_proxy_unset() {
  echo " - Unsetting HTTP(S) proxies if set."
  unset NO_PROXY
  unset HTTP_PROXY
  unset HTTPS_PROXY
  unset no_proxy
  unset http_proxy
  unset https_proxy
}


function f_amulet_no_teardown() {
# Hack up the makefile to prevent amulet from tearing down enviro on failured
  cd $CO_DIR
  $OSCI_ROOT/run/tools/find_replace.py -yd -i Makefile -o Makefile -f "juju test -v" -r "juju test --set-e -v" ||:
}


function f_amulet_set_timeout() {
# Override known common timeout values in legacy amulet test runners
  $OSCI_ROOT/run/tools/find_replace.py -yd -i Makefile -o Makefile -f "--timeout 900" -r "--timeout 2700" ||:
  $OSCI_ROOT/run/tools/find_replace.py -yd -i Makefile -o Makefile -f "--timeout 1200" -r "--timeout 2700" ||:
}


function f_gather_package_info() {
  echo " . Gathering package info"
  tmp_file="$(mktemp)"
  cat > "$tmp_file" << "--EOF"
  amulet
  charm
  charm-tools
  juju
  juju-1.25
  juju-2.0
  juju-core
  juju-deployer
  lxc
  lxd
  mojo
  python-amqp
  python-ceilometerclient
  python-cinderclient
  python-glanceclient
  python-heatclient
  python-jujuclient
  python-keystoneclient
  python-neutronclient
  python-novaclient
  python-openstackclient
  python-swiftclient
  python-tox
  linux-virtual
--EOF

  for line in $(cat $tmp_file); do
    str="${str}${line}|"
    apt-cache policy $line 2>&1 | tee -a $WORKSPACE/pkg_info.txt
  done
  match=$(echo $str | head -c -2)
  eval "dpkg -l | egrep '$match'" 2>&1 | tee -a $WORKSPACE/pkg_info.txt
  rm -vf $tmp_file
}


function f_get_my_nova_instances() {
# Get nova instances for the current juju environment
  if [[ "${JENV_NAME^^}" != *MAAS* ]]; then
    # export my_instances=$(timeout 60 openstack server list | grep "juju\-${JENV_NAME}\-machine" ||:)

    # XXX: Deprecate or fixme.  Dangerous as is due to instance name diffs from Juju 1 to 2.
    export my_instances="DeprecateOrFixMe"
    exit 1
  fi
}


function f_juju_bootstrap() {
  # Repo should be present before now, but if it's not, get it!
  stat $CTI_CODIR/tox.ini ||\
      f_tool_refresh_cti

  f_novarc_undercloud
  export BOOTSTRAP_CONSTRAINTS="virt-type=kvm cores=4 mem=3G"
  (
    cd $CTI_CODIR
    retry_command ./juju-openstack-controller-example.sh
  )
}


function f_juju_bootstrap_maas() {
  # Repo should be present before now, but if it's not, get it!
  stat $CTI_CODIR/tox.ini ||\
      f_tool_refresh_cti

  export CLOUD_NAME="ruxton-maas"
  export CONTROLLER_NAME="ruxton-maas-${NODE_NAME}"
  export BOOTSTRAP_CONSTRAINTS="arch=$ARCH tags=uosci"
  export MODEL_CONSTRAINTS="$BOOTSTRAP_CONSTRAINTS"
  (
    cd $CTI_CODIR
    retry_command ./juju-maas-controller-example.sh
  )
}


function f_juju_stat() {
  juju status --format tabular &> $WORKSPACE/juju-status${1}.txt
  juju status --format yaml    &> $WORKSPACE/juju-status${1}.yaml.txt
  juju status --format json    &> $WORKSPACE/juju-status${1}.json.txt

  juju controllers             &> $WORKSPACE/juju-controllers${1}.txt
  juju models                  &> $WORKSPACE/juju-controllers${1}.txt
  cat $WORKSPACE/juju-status${1}.txt
}


function f_lolo_log_puller() {
    f_line $FUNCNAME
    f_activate_botcontrol_tools_virtualenv
    $UOSCI_BOT_CONTROL_ROOT/tools/lolo-log-puller $WORKSPACE/logs
}


function f_juju_destroy_controller() {
  # Repo should be present before now, but if it's not, get it!
  stat $CTI_CODIR/tox.ini ||\
      f_tool_refresh_cti

  f_novarc_undercloud
  (
    cd $CTI_CODIR
    retry_command ./juju-destroy-controller-and-models-example.sh -y
  )
}


function f_juju_destroy_model() {
  # Repo should be present before now, but if it's not, get it!
  stat $CTI_CODIR/tox.ini ||\
      f_tool_refresh_cti

  f_novarc_undercloud
  (
    cd $CTI_CODIR
    retry_command ./juju-destroy-model-example.sh -y
    retry_command ./juju-destroy-zaza-models.sh -y
  )
}


function f_activate_clients() {
  deactivate ||:

  # Tool should be present before now, but if it's not, get it!
  stat $CTI_CODIR/tox.ini ||\
      f_tool_refresh_cti

  # Venv may have already been built, but if not, build it!
  stat $CTI_CODIR/.tox ||\
  (
      cd $CTI_CODIR
      tox -e clients
  )
  . $CTI_CODIR/clientsrc
}


function f_activate_botcontrol_tools_virtualenv() {
  deactivate ||:

  # Tool should be present before now, but if it's not, get it!
  stat $UOSCI_BOT_CONTROL_ROOT/tox.ini ||\
      f_tool_refresh_bot_control

  # Venv may have already been built, but if not, build it!
  stat $UOSCI_BOT_CONTROL_ROOT/.tox/port-cleanup ||\
  (
      cd $UOSCI_BOT_CONTROL_ROOT/tools
      tox -e port-cleanup
  )
  . $UOSCI_BOT_CONTROL_ROOT/tools/.tox/port-cleanup/bin/activate
}


function f_tmp_cleanup(){
# Clean up tmp dir items as the jenkins user
  f_line "Cleaning up /tmp (Safe to ignore file/dir permission errors)"
  time find /tmp -type f -user jenkins -writable -readable -mmin +60 | xargs rm -rf ||:
  time find /tmp -type d -user jenkins -writable -readable -mmin +60 | xargs rm -rf ||:
}


function f_co_cleanup(){
# Clean up checkout dir items as the jenkins user
    if [[ -d "$CO_DIR" ]] && [[ -n "$(echo $CO_DIR | grep '/var/lib/jenkins')" ]]; then
        f_line "Cleaning up checkout dir $CO_DIR"
        rm -rf $CO_DIR
    fi
}


function f_line() {
# Useful to announce in the console log
  echo "====>  ${1}  <======================="
}


function f_skip_juju_foo() {
  export SKIP_JUJU_FOO="true"
}


# Avoid environment teardown if run outside of a Jenkins environment
if [[ -z "$BUILD_URL" ]]; then
    export NO_POST_DESTROY="true"
fi

# Charm store timeout value, consumed by libcharmstore
export CS_API_TIMEOUT=60

# Charm-tools work-around
export USER="$(whoami)"

f_line "osci_job_common complete"
