#!/bin/bash -e
#  Common functions for UOSCI layer ci, charm build, etc.

function f_layerci_use_built_asset_if_relevant(){
# Collect and extract built asset archive if relevant.
  if f_layerci_in_effect; then
    f_layerci_built_asset_var_sanity_checks
    f_layerci_scp_get_built_asset_archive_file "$BUILT_ASSET_ARTIFACT_ARCHIVE_NAME"
    f_layerci_validate_retrieved_asset_archive_file "$BUILT_ASSET_ARTIFACT_ARCHIVE_NAME" "$BUILT_ASSET_SHA1"
    f_layerci_extract_asset_archive_file "$BUILT_ASSET_ARTIFACT_ARCHIVE_NAME" "$BUILT_ASSET_BASE_NAME"
  fi
}


function f_layerci_in_effect(){
# Use env vars to determent if layerci (built assets) are in effect.
# Such env vars will exist via job parameters if relevant.
  if [[ -n "$BUILT_ASSET_ARTIFACT_ARCHIVE_NAME" ]] &&\
     [[ -n "$BUILT_ASSET_BASE_NAME" ]] &&\
     [[ -n "$BUILT_ASSET_SHA1" ]]; then
    return 0
  else
    # Not in effect.
    return 1
  fi
}


function f_layerci_built_asset_var_sanity_checks(){
# Make sure the built asset name matches the base name of the job.
# Fail and run far and fast if not.
  f_line "Sanity check on BUILT_ASSET_* values"
  if [[ -n "$BUILT_ASSET_BASE_NAME" ]] && [[ -n "$BASE_NAME" ]] &&\
     [[ "$BUILT_ASSET_BASE_NAME" != "$BASE_NAME" ]]; then
    echo " ! BUILT_ASSET_BASE_NAME does not match BASE_NAME.  Bailing!"
    env | grep BUILT_ASSET | sort
    exit 1
  else
    echo " . OK: BUILT_ASSET_BASE_NAME matches BASE_NAME."
    env | grep BUILT_ASSET | sort
  fi
}


function f_layerci_scp_get_built_asset_archive_file(){
# Retrieve a built asset archive file from shared artifact storage and
# place it in the workspace asset dir.  Takes 1 parameter: the archive filename.
  ARCHIVE_NAME="$1"
  f_line "Retrieving $ARCHIVE_NAME from artifact storage"
  if [[ -z "$ARCHIVE_NAME" ]]; then
    echo " ! Missing required parameter:  built asset artifact archive name"
    exit 1
  fi
  mkdir -p $WORKSPACE/asset
  scp -p jenkins@${OSCI_MASTER_IP}:/mnt/jenkins-ssh-artifact-storage/charm_build/$ARCHIVE_NAME $WORKSPACE/asset/$ARCHIVE_NAME
  stat $WORKSPACE/asset/$ARCHIVE_NAME
}


function f_layerci_validate_retrieved_asset_archive_file(){
# Validate hash sum of the retrieved archive file against the hash value passed
# from the earlier build job.
  ARCHIVE_NAME="$1"
  f_line "Validating $ARCHIVE_NAME"
  if [[ -z "$ARCHIVE_NAME" ]]; then
    echo " ! Missing required parameter:  built asset artifact archive name"
    exit 1
  fi
  ARCHIVE_SHA1="$2"
  if [[ -z "$ARCHIVE_SHA1" ]]; then
    echo " ! Missing required parameter:  artifact archive sha1"
    exit 1
  fi

  RETRIEVED_SHA1="$(cd $WORKSPACE/asset; sha1sum $ARCHIVE_NAME)"
  if [[ "$ARCHIVE_SHA1" == "$RETRIEVED_SHA1" ]]; then
    echo " . SHA1 check ok for: $WORKSPACE/asset/$ARCHIVE_NAME ($RETRIEVED_SHA1)"
  else
    echo " ! SHA1 check failed for: $WORKSPACE/asset/$ARCHIVE_NAME"
    echo "   Actual:   $RETRIEVED_SHA1"
    echo "   Expected: $ARCHIVE_SHA1"
    exit 1
  fi
}


function f_layerci_extract_asset_archive_file(){
# Extract/expand an asset archive file into the asset dir.
  f_line "Extracting $ARCHIVE_NAME to $WORKSPACE/asset/$ASSET_NAME"
  ARCHIVE_NAME="$1"
  ASSET_NAME="$2"
  if [[ -z "$ARCHIVE_NAME" ]]; then
    echo " ! Missing required parameter:  built asset artifact archive name"
    exit 1
  fi
  if [[ -z "$ASSET_NAME" ]]; then
    echo " ! Missing required parameter:  asset name (ie. charm name)"
    exit 1
  fi
  mkdir -p $WORKSPACE/asset/$ASSET_NAME
  tar -jxvf $WORKSPACE/asset/$ARCHIVE_NAME -C $WORKSPACE/asset/$ASSET_NAME
  ls -al $WORKSPACE/asset/$ASSET_NAME
  export BUILT_ASSET_DIR=$WORKSPACE/asset/$ASSET_NAME
}


function f_layerci_get_asset_type(){
# Takes 1 required parameter: the check out path.
# Sets env vars to be consumed by jobs.
  TMP_DIR="$1"
  if [[ -z "$TMP_DIR" ]]; then
    echo " ! Missing required parameter:  checkout dir"
    exit 1
  fi
  f_line "Checking asset type of $TMP_DIR"

  ${UOSCI_BOT_CONTROL_ROOT}/tools/charm-what $TMP_DIR  # Exits non-zero if unable to determine type
  export OSCI_ASSET_TYPE="$(${UOSCI_BOT_CONTROL_ROOT}/tools/charm-what $TMP_DIR)"

  if [[ -z "$OSCI_ASSET_TYPE" ]]; then
    # Unlikely to hit this; trap anyway.
    echo " ! Unable to determine asset type for $TMP_DIR"
    exit 1
  fi
}


function f_layerci_build_charm(){
# Build asset (charm) based on checkout dir and base name env vars,
# and set the built asset dir env var to be consumed by jobs.
  if [[ -z "$CO_DIR" ]]; then
    echo " ! Missing required CO_DIR env var"
    exit 1
  fi
  if [[ -z "$BASE_NAME" ]]; then
    echo " ! Missing required BASE_NAME env var"
    exit 1
  fi
  f_line "Building $CO_DIR ($BASE_NAME)"

  # Build the asset
  f_proxy_if_proxy
  . $UOSCI_RELEASE_TOOLS_ROOT/build-charm $CO_DIR
  f_proxy_unset

  # Confirm built asset location
  if [[ -z "$BUILT_ASSET_DIR" ]]; then
    echo " ! Unable to determine BUILT_ASSET_DIR"
    exit 1
  fi
  echo " . BUILT_ASSET_DIR: $BUILT_ASSET_DIR"
  du -hsc $BUILT_ASSET_DIR

  # Inject repo summary into the built asset
  $UOSCI_RELEASE_TOOLS_ROOT/generate-repo-info $CO_DIR &> $BUILT_ASSET_DIR/repo-info
}


function f_layerci_archive_built_asset(){
# Assumes the asset (charm) has already been built.
# Sets the properties files for later jobs to consume.
  f_line "Archiving built asset $BUILT_ASSET_DIR as artifact"
  if [[ -z "$BUILT_ASSET_DIR" ]]; then
    echo " ! Missing required env var BUILT_ASSET_DIR"
    exit 1
  fi
  if [[ ! -d "$BUILT_ASSET_DIR" ]]; then
    echo " ! BUILT_ASSET_DIR does not exist"
    exit 1
  fi

  # Create archive and make sure it exists
  # Builds always go in workspace/build/
  mkdir -p $WORKSPACE/build
  export BUILT_ASSET_ARTIFACT_ARCHIVE_NAME="${BASE_NAME}-${JOB_NAME}-${BUILD_NUMBER}.tar.bz2"
  tar -C $BUILT_ASSET_DIR -cjvf $WORKSPACE/build/$BUILT_ASSET_ARTIFACT_ARCHIVE_NAME .
  ls -alh $WORKSPACE/build/$BUILT_ASSET_ARTIFACT_ARCHIVE_NAME

  # Create properties files for parameterized job consumption. These env vars
  # are critical to pass along for layer ci awareness in other jobs.
  echo "BUILT_ASSET_ARTIFACT_ARCHIVE_NAME=$BUILT_ASSET_ARTIFACT_ARCHIVE_NAME" > $WORKSPACE/charm_build.properties
  echo "BUILT_ASSET_BASE_NAME=$BASE_NAME" >> $WORKSPACE/charm_build.properties
  echo "BUILT_ASSET_SHA1=$(cd $WORKSPACE/build; sha1sum $BUILT_ASSET_ARTIFACT_ARCHIVE_NAME)" >> $WORKSPACE/charm_build.properties
}


function f_layerci_build_and_archive_if_necessary() {
# Detect asset (charm/layer/interface) type, build and archive if relevant.
# Announce detection for clarity.  Expects CO_DIR and OSCI_ASSET_TYPE
# env vars to already be set.
  case "$OSCI_ASSET_TYPE" in
    "charm (built)")
        # Change nothing at this time
        echo " . Proceeding with $CO_DIR as an already-built charm [no build necessary]"
        ;;
    "charm (classic)")
        # Change nothing at this time
        echo " . Proceeding with $CO_DIR as a classic (non-layered) charm [no build necessary]"
        ;;
    "charm (source)")
        # Build and prepare jenkins workspace artifact for copying by other jobs
        echo " . Proceeding with $CO_DIR as a source charm [need to build]"
        f_layerci_build_charm $CO_DIR
        f_layerci_archive_built_asset
        ;;
    "layer")
        # Build and prepare jenkins workspace artifact for copying by other jobs
        echo " . Proceeding with $CO_DIR as a layer [need to build]"
        f_layerci_build_charm $CO_DIR
        f_layerci_archive_built_asset
        ;;
    "interface")
        # Change nothing at this time
        echo " . Proceeding with $CO_DIR as an interface [no build necessary]"
        ;;
    *)
        echo " ! Unknown asset type"
        exit 1
        ;;
  esac
}


function f_layerci_fail_if_not_deployable_type(){
# This is used in front of deployment tests, such as charm single,
# amulet and such, to not waste time bootstrapping and attempting
# to deploy an unbuilt layer or interface. Those jobs should never
# attempt that. This makes sure and indicates if/when they do try.

  f_layerci_check_and_announce_asset_type
  case "$OSCI_ASSET_TYPE" in
    "charm (built)"|"charm (classic)")
        # OK
        :
        ;;
    "charm (source)"|"layer"|"interface")
        # Bail!
        echo " ! Refusing to attempt a deploy test with a not-deployable asset: $OSCI_ASSET_TYPE"
        exit 1
        ;;
    *)
        #
        echo " ! Unknown asset type"
        exit 1
        ;;
  esac
}


function f_layerci_build_if_necessary() {
    f_layerci_check_and_announce_asset_type
    case "$OSCI_ASSET_TYPE" in
        "charm (built)"|"charm (classic)")
            # OK
            :
            ;;
        "charm (source)")
            f_layerci_build_charm
            ;;
        *)
            #
            echo " ! Asset type not supported in this call: $OSCI_ASSET_TYPE"
            exit 1
        ;;
    esac
}


function f_layerci_check_and_announce_asset_type() {
# Check (or re-check) and announce asset type via charm-what.
  f_layerci_get_asset_type ${BUILT_ASSET_DIR-$CO_DIR}
  case "$OSCI_ASSET_TYPE" in
    "charm (built)"|"charm (classic)"|"charm (source)"|"layer"|"interface")
        # OK
        :
        ;;
    *)
        echo " ! Unknown asset type"
        exit 1
        ;;
  esac
}
